#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""mythril.py: Bug hunting on the Ethereum blockchain

   http://www.github.com/ConsenSys/mythril
"""

import logging, coloredlogs
import json
import sys
import argparse

# logging.basicConfig(level=logging.DEBUG)

from mythril.exceptions import CriticalError, AddressNotFoundError
from mythril.mythril import Mythril
from mythril.version import VERSION
import mythril.support.signatures as sigs


def exit_with_error(format_, message):
    if format_ == 'text' or format_ == 'markdown':
        print(message)
    else:
        result = {'success': False, 'error': str(message), 'issues': []}
        print(json.dumps(result))
    sys.exit()


def main():
    parser = argparse.ArgumentParser(description='Security analysis of Ethereum smart contracts')
    parser.add_argument("solidity_file", nargs='*')

    commands = parser.add_argument_group('commands')
    commands.add_argument('-g', '--graph', help='generate a control flow graph')
    commands.add_argument('-V', '--version', action='store_true',
                          help='print the Mythril version number and exit')
    commands.add_argument('-x', '--fire-lasers', action='store_true',
                          help='detect vulnerabilities, use with -c, -a or solidity file(s)')
    commands.add_argument('-t', '--truffle', action='store_true',
                          help='analyze a truffle project (run from project dir)')
    commands.add_argument('-d', '--disassemble', action='store_true', help='print disassembly')
    commands.add_argument('-j', '--statespace-json', help='dumps the statespace json', metavar='OUTPUT_FILE')

    inputs = parser.add_argument_group('input arguments')
    inputs.add_argument('-c', '--code', help='hex-encoded bytecode string ("6060604052...")', metavar='BYTECODE')
    inputs.add_argument('-f', '--codefile', help='file containing hex-encoded bytecode string', 
            metavar='BYTECODEFILE', type=argparse.FileType('r'))
    inputs.add_argument('-a', '--address', help='pull contract from the blockchain', metavar='CONTRACT_ADDRESS')
    inputs.add_argument('-l', '--dynld', action='store_true', help='auto-load dependencies from the blockchain')

    outputs = parser.add_argument_group('output formats')
    outputs.add_argument('-o', '--outform', choices=['text', 'markdown', 'json'], default='text',
                         help='report output format', metavar='<text/markdown/json>')
    outputs.add_argument('--verbose-report', action='store_true', help='Include debugging information in report')

    database = parser.add_argument_group('local contracts database')
    database.add_argument('-s', '--search', help='search the contract database', metavar='EXPRESSION')
    database.add_argument('--leveldb-dir', help='specify leveldb directory for search or direct access operations', metavar='LEVELDB_PATH')

    utilities = parser.add_argument_group('utilities')
    utilities.add_argument('--hash', help='calculate function signature hash', metavar='SIGNATURE')
    utilities.add_argument('--storage', help='read state variables from storage index, use with -a',
                           metavar='INDEX,NUM_SLOTS,[array] / mapping,INDEX,[KEY1, KEY2...]')
    utilities.add_argument('--solv',
                           help='specify solidity compiler version. If not present, will try to install it (Experimental)',
                           metavar='SOLV')
    utilities.add_argument('--contract-hash-to-address', help='returns corresponding address for a contract address hash', metavar='SHA3_TO_LOOK_FOR')

    options = parser.add_argument_group('options')
    options.add_argument('-m', '--modules', help='Comma-separated list of security analysis modules', metavar='MODULES')
    options.add_argument('--max-depth', type=int, default=22, help='Maximum recursion depth for symbolic execution')

    options.add_argument('--strategy', choices=['dfs', 'bfs', 'naive-random', 'weighted-random'],
                         default='dfs', help='Symbolic execution strategy')
    options.add_argument('--max-transaction-count', type=int, default=3, help='Maximum number of transactions issued by laser')

    options.add_argument('--execution-timeout', type=int, default=600, help="The amount of seconds to spend on symbolic execution")
    options.add_argument('--create-timeout', type=int, default=10, help="The amount of seconds to spend on "
                                                                        "the initial contract creation")
    options.add_argument('--solc-args', help='Extra arguments for solc')
    options.add_argument('--phrack', action='store_true', help='Phrack-style call graph')
    options.add_argument('--enable-physics', action='store_true', help='enable graph physics simulation')
    options.add_argument('-v', type=int, help='log level (0-2)', metavar='LOG_LEVEL')
    options.add_argument('-q', '--query-signature', action='store_true', help='Lookup function signatures through www.4byte.directory')

    rpc = parser.add_argument_group('RPC options')
    rpc.add_argument('-i', action='store_true', help='Preset: Infura Node service (Mainnet)')
    rpc.add_argument('--rpc', help='custom RPC settings', metavar='HOST:PORT / ganache / infura-[network_name]')
    rpc.add_argument('--rpctls', type=bool, default=False, help='RPC connection over TLS')

    # Get config values

    args = parser.parse_args()

    if args.version:
        if args.outform == 'json':
            print(json.dumps({'version_str': VERSION}))
        else:
            print("Mythril version {}".format(VERSION))
        sys.exit()

    # Parse cmdline args

    if not (args.search or args.hash or args.disassemble or args.graph or args.fire_lasers
            or args.storage or args.truffle or args.statespace_json or args.contract_hash_to_address):
        parser.print_help()
        sys.exit()


    if args.v:
        if 0 <= args.v < 3:
            coloredlogs.install(
                fmt='%(name)s[%(process)d] %(levelname)s %(message)s',
                level=[logging.NOTSET, logging.INFO, logging.DEBUG][args.v]
            )
        else:
            exit_with_error(args.outform, "Invalid -v value, you can find valid values in usage")

    if args.query_signature:
        if sigs.ethereum_input_decoder == None:
            exit_with_error(args.outform, "The --query-signature function requires the python package ethereum-input-decoder")

    # -- commands --
    if args.hash:
        print(Mythril.hash_for_function_signature(args.hash))
        sys.exit()

    try:
        # the mythril object should be our main interface
        # infura = None, rpc = None, rpctls = None
        # solc_args = None, dynld = None, max_recursion_depth = 12):

        mythril = Mythril(solv=args.solv, dynld=args.dynld,
                          solc_args=args.solc_args,
                          enable_online_lookup=args.query_signature)
        if args.dynld and not (args.rpc or args.i):
            mythril.set_api_from_config_path()

        if args.address:
            # Establish RPC connection if necessary
            if args.i:
                mythril.set_api_rpc_infura()
            elif args.rpc:
                mythril.set_api_rpc(rpc=args.rpc, rpctls=args.rpctls)
            elif not args.dynld:
                mythril.set_api_rpc_localhost()
        elif args.search or args.contract_hash_to_address:
            # Open LevelDB if necessary
            mythril.set_api_leveldb(mythril.leveldb_dir if not args.leveldb_dir else args.leveldb_dir)

        if args.search:
            # Database search ops
            mythril.search_db(args.search)
            sys.exit()

        if args.contract_hash_to_address:
            # search corresponding address
            try:
                mythril.contract_hash_to_address(args.contract_hash_to_address)
            except AddressNotFoundError:
                print("Address not found.")

            sys.exit()

        if args.truffle:
            try:
                # not really pythonic atm. needs refactoring
                mythril.analyze_truffle_project(args)
            except FileNotFoundError:
                print(
                    "Build directory not found. Make sure that you start the analysis from the project root, and that 'truffle compile' has executed successfully.")
            sys.exit()

        # Load / compile input contracts
        address = None

        if args.code:
            # Load from bytecode
            address, _ = mythril.load_from_bytecode(args.code)
        elif args.codefile:
            bytecode = ''.join([l.strip() for l in args.codefile if len(l.strip()) > 0])
            address, _ = mythril.load_from_bytecode(bytecode)
        elif args.address:
            # Get bytecode from a contract address
            address, _ = mythril.load_from_address(args.address)
        elif args.solidity_file:
            # Compile Solidity source file(s)
            if args.graph and len(args.solidity_file) > 1:
                exit_with_error(args.outform,
                                "Cannot generate call graphs from multiple input files. Please do it one at a time.")
            address, _ = mythril.load_from_solidity(args.solidity_file)  # list of files
        else:
            exit_with_error(args.outform,
                            "No input bytecode. Please provide EVM code via -c BYTECODE, -a ADDRESS, or -i SOLIDITY_FILES")

        # Commands

        if args.storage:
            if not args.address:
                exit_with_error(args.outform,
                                "To read storage, provide the address of a deployed contract with the -a option.")

            storage = mythril.get_state_variable_from_storage(address=address,
                                                              params=[a.strip() for a in args.storage.strip().split(",")])
            print(storage)

        elif args.disassemble:
            easm_text = mythril.contracts[0].get_easm() # or mythril.disassemble(mythril.contracts[0])
            sys.stdout.write(easm_text)

        elif args.graph or args.fire_lasers:
            if not mythril.contracts:
                exit_with_error(args.outform, "input files do not contain any valid contracts")

            if args.graph:
                html = mythril.graph_html(strategy=args.strategy, contract=mythril.contracts[0], address=address,
                                          enable_physics=args.enable_physics, phrackify=args.phrack,
                                          max_depth=args.max_depth, execution_timeout=args.execution_timeout,
                                          create_timeout=args.create_timeout)

                try:
                    with open(args.graph, "w") as f:
                        f.write(html)
                except Exception as e:
                    exit_with_error(args.outform, "Error saving graph: " + str(e))

            else:
                report = mythril.fire_lasers(strategy=args.strategy, address=address,
                                             modules=[m.strip() for m in args.modules.strip().split(",")] if args.modules else [],
                                             verbose_report=args.verbose_report,
                                             max_depth=args.max_depth, execution_timeout=args.execution_timeout,
                                             create_timeout=args.create_timeout,
                                             max_transaction_count=args.max_transaction_count)
                outputs = {
                    'json': report.as_json(),
                    'text': report.as_text(),
                    'markdown': report.as_markdown()
                }
                print(outputs[args.outform])

        elif args.statespace_json:

            if not mythril.contracts:
                exit_with_error(args.outform, "input files do not contain any valid contracts")

            statespace = mythril.dump_statespace(strategy=args.strategy, contract=mythril.contracts[0], address=address,
                                                 max_depth=args.max_depth, execution_timeout=args.execution_timeout,
                                                 create_timeout=args.create_timeout)

            try:
                with open(args.statespace_json, "w") as f:
                    json.dump(statespace, f)
            except Exception as e:
                exit_with_error(args.outform, "Error saving json: " + str(e))

        else:
            parser.print_help()

    except CriticalError as ce:
        exit_with_error(args.outform, str(ce))


if __name__ == "__main__":
    main()



from mythril.disassembler.disassembly import Disassembly
import logging
import re


class DynLoader:

    def __init__(self, eth):
        self.eth = eth
        self.storage_cache = {}

    def read_storage(self, contract_address, index):

        try:
            contract_ref = self.storage_cache[contract_address]
            data = contract_ref[index]

        except KeyError:

            self.storage_cache[contract_address] = {}

            data = self.eth.eth_getStorageAt(contract_address, position=index, block='latest')

            self.storage_cache[contract_address][index] = data

        except IndexError:

            data = self.eth.eth_getStorageAt(contract_address, position=index, block='latest')

            self.storage_cache[contract_address][index] = data

        return data

    def dynld(self, contract_address, dependency_address):

        logging.info("Dynld at contract " + contract_address + ": " + dependency_address)

        m = re.match(r'^(0x[0-9a-fA-F]{40})$', dependency_address)

        if m:
            dependency_address = m.group(1)

        else:
            return None

        logging.info("Dependency address: " + dependency_address)

        code = self.eth.eth_getCode(dependency_address)

        if code == "0x":
            return None
        else:
            return Disassembly(code)



import re
from z3 import *
from mythril.analysis.ops import VarType
from mythril.analysis import solver
from mythril.analysis.report import Issue
from mythril.analysis.swc_data import TIMESTAMP_DEPENDENCE, PREDICTABLE_VARS_DEPENDENCE
from mythril.exceptions import UnsatError
import logging

'''
MODULE DESCRIPTION:

Check for CALLs that send >0 Ether as a result of computation based on predictable variables such as
block.coinbase, block.gaslimit, block.timestamp, block.number

TODO:
- block.blockhash(block.number-1)
- block.blockhash(some_block_past_256_blocks_from_now)==0
- external source of random numbers (e.g. Oraclize)
'''


def execute(statespace):

    logging.debug("Executing module: DEPENDENCE_ON_PREDICTABLE_VARS")

    issues = []

    for call in statespace.calls:

        if "callvalue" in str(call.value):
            logging.debug("[DEPENDENCE_ON_PREDICTABLE_VARS] Skipping refund function")
            continue

        # We're only interested in calls that send Ether

        if call.value.type == VarType.CONCRETE and call.value.val == 0:
            continue

        address = call.state.get_current_instruction()['address']

        description = "In the function `" + call.node.function_name + "` "
        description += "the following predictable state variables are used to determine Ether recipient:\n"

        # First check: look for predictable state variables in node & call recipient constraints

        vars = ["coinbase", "gaslimit", "timestamp", "number"]

        found = []
        for var in vars:
            for constraint in call.node.constraints + [call.to]:
                if var in str(constraint):
                    found.append(var)

        if len(found):
            for item in found:
                description += "- block.{}\n".format(item)
            if solve(call):
                swc_type = TIMESTAMP_DEPENDENCE if item == 'timestamp' else PREDICTABLE_VARS_DEPENDENCE
                issue = Issue(contract=call.node.contract_name, function=call.node.function_name, address=address,
                              swc_id=swc_type, title="Dependence on predictable environment variable",
                              _type="Warning", description=description)
                issues.append(issue)

        # Second check: blockhash

        for constraint in call.node.constraints + [call.to]:
            if "blockhash" in str(constraint):
                description = "In the function `" + call.node.function_name + "` "
                if "number" in str(constraint):
                    m = re.search(r'blockhash\w+(\s-\s(\d+))*', str(constraint))
                    if m and solve(call):

                        found = m.group(1)

                        if found:  # block.blockhash(block.number - N)
                            description += "predictable expression 'block.blockhash(block.number - " + m.group(2) + \
                                ")' is used to determine Ether recipient"
                            if int(m.group(2)) > 255:
                                description += ", this expression will always be equal to zero."
                        elif "storage" in str(constraint):  # block.blockhash(block.number - storage_0)
                            description += "predictable expression 'block.blockhash(block.number - " + \
                                           "some_storage_var)' is used to determine Ether recipient"
                        else:  # block.blockhash(block.number)
                            description += "predictable expression 'block.blockhash(block.number)'" + \
                                           " is used to determine Ether recipient"
                            description += ", this expression will always be equal to zero."

                        issue = Issue(contract=call.node.contract_name, function=call.node.function_name,
                                      address=address, title="Dependence on predictable variable",
                                      _type="Warning", description=description, swc_id=PREDICTABLE_VARS_DEPENDENCE)
                        issues.append(issue)
                        break
                else:
                    r = re.search(r'storage_([a-z0-9_&^]+)', str(constraint))
                    if r:  # block.blockhash(storage_0)

                        '''
                        We actually can do better here by adding a constraint blockhash_block_storage_0 == 0
                        and checking model satisfiability. When this is done, severity can be raised
                        from 'Informational' to 'Warning'.
                        Checking that storage at given index can be tainted is not necessary, since it usually contains
                        block.number of the 'commit' transaction in commit-reveal workflow.
                        '''

                        index = r.group(1)
                        if index and solve(call):
                            description += 'block.blockhash() is calculated using a value from storage ' \
                                           'at index {}'.format(index)
                            issue = Issue(contract=call.node.contract_name, function=call.node.function_name,
                                          address=address, title="Dependence on predictable variable",
                                          _type="Informational", description=description, swc_id=PREDICTABLE_VARS_DEPENDENCE)
                            issues.append(issue)
                            break
    return issues


def solve(call):
    try:
        model = solver.get_model(call.node.constraints)
        logging.debug("[DEPENDENCE_ON_PREDICTABLE_VARS] MODEL: " + str(model))

        for decl in model.decls():
            logging.debug("[DEPENDENCE_ON_PREDICTABLE_VARS] main model: %s = 0x%x" % (decl.name(), model[decl].as_long()))
        return True

    except UnsatError:
        logging.debug("[DEPENDENCE_ON_PREDICTABLE_VARS] no model found")
        return False



import hashlib
import json
import operator
from jinja2 import PackageLoader, Environment


class Issue:

    def __init__(self, contract, function, address, swc_id, title, _type="Informational", description="", debug=""):

        self.title = title
        self.contract = contract
        self.function = function
        self.address = address
        self.description = description
        self.type = _type
        self.debug = debug
        self.swc_id = swc_id
        self.filename = None
        self.code = None
        self.lineno = None


    @property
    def as_dict(self):

        issue = {'title': self.title, 'swc_id': self.swc_id, 'contract': self.contract, 'description': self.description,
                 'function': self.function, 'type': self.type, 'address': self.address, 'debug': self.debug}

        if self.filename and self.lineno:
            issue['filename'] = self.filename
            issue['lineno'] = self.lineno

        if self.code:
            issue['code'] = self.code

        return issue

    def add_code_info(self, contract):
        if self.address:
            codeinfo = contract.get_source_info(self.address, constructor=(self.function == 'constructor'))
            self.filename = codeinfo.filename
            self.code = codeinfo.code
            self.lineno = codeinfo.lineno


class Report:
    environment = Environment(loader=PackageLoader('mythril.analysis'), trim_blocks=True)

    def __init__(self, verbose=False):
        self.issues = {}
        self.verbose = verbose
        pass

    def sorted_issues(self):
        issue_list = [issue.as_dict for key, issue in self.issues.items()]
        return sorted(issue_list, key=operator.itemgetter('address', 'title'))

    def append_issue(self, issue):
        m = hashlib.md5()
        m.update((issue.contract + str(issue.address) + issue.title).encode('utf-8'))
        self.issues[m.digest()] = issue

    def as_text(self):
        name = self._file_name()
        template = Report.environment.get_template('report_as_text.jinja2')
        return template.render(filename=name, issues=self.sorted_issues(), verbose=self.verbose)

    def as_json(self):
        result = {'success': True, 'error': None, 'issues': self.sorted_issues()}
        return json.dumps(result, sort_keys=True)

    def as_markdown(self):
        filename = self._file_name()
        template = Report.environment.get_template('report_as_markdown.jinja2')
        return template.render(filename=filename, issues=self.sorted_issues(), verbose=self.verbose)

    def _file_name(self):
        if len(self.issues.values()) > 0:
            return list(self.issues.values())[0].filename



import sys
import re
from ethereum.opcodes import opcodes
from mythril.ether import util


regex_PUSH = re.compile('^PUSH(\d*)$')

# Additional mnemonic to catch failed assertions

opcodes[254] = ['ASSERT_FAIL', 0, 0, 0]


def instruction_list_to_easm(instruction_list):
    easm = ""

    for instruction in instruction_list:

        easm += str(instruction['address']) + " " + instruction['opcode']

        if 'argument' in instruction:
            easm += " " + instruction['argument']

        easm += "\n"

    return easm


def easm_to_instruction_list(easm):

    regex_CODELINE = re.compile('^([A-Z0-9]+)(?:\s+([0-9a-fA-Fx]+))?$')

    instruction_list = []

    codelines = easm.split('\n')

    for line in codelines:

        m = re.search(regex_CODELINE, line)

        if not m:
            # Invalid code line
            continue

        instruction = {'opcode': m.group(1)}

        if m.group(2):
            instruction['argument'] = m.group(2)[2:]

        instruction_list.append(instruction)

    return instruction_list


def get_opcode_from_name(name):

    for opcode, value in opcodes.items():

        if name == value[0]:

            return opcode

    raise RuntimeError("Unknown opcode")


def find_opcode_sequence(pattern, instruction_list):
    match_indexes = []

    pattern_length = len(pattern)

    for i in range(0, len(instruction_list) - pattern_length + 1):

        if instruction_list[i]['opcode'] in pattern[0]:

            matched = True

            for j in range(1, len(pattern)):

                if not (instruction_list[i + j]['opcode'] in pattern[j]):
                    matched = False
                    break

            if matched:
                match_indexes.append(i)

    return match_indexes


def disassemble(bytecode):

    instruction_list = []
    addr = 0

    length = len(bytecode)

    if "bzzr" in str(bytecode[-43:]):
        # ignore swarm hash
        length -= 43

    while addr < length:

        instruction = {'address': addr}

        try:
            if sys.version_info > (3, 0):
                opcode = opcodes[bytecode[addr]]
            else:
                opcode = opcodes[ord(bytecode[addr])]

        except KeyError:

            # invalid opcode
            instruction_list.append({'address': addr, 'opcode': "INVALID"})
            addr += 1
            continue

        instruction['opcode'] = opcode[0]

        m = re.search(regex_PUSH, opcode[0])

        if m:
            argument = bytecode[addr+1:addr+1+int(m.group(1))]
            instruction['argument'] = "0x" + argument.hex()

            addr += int(m.group(1))


        instruction_list.append(instruction)

        addr += 1

    return instruction_list


def assemble(instruction_list):

    bytecode = b""

    for instruction in instruction_list:

        try:
            opcode = get_opcode_from_name(instruction['opcode'])
        except RuntimeError:
            opcode = 0xbb

        bytecode += opcode.to_bytes(1, byteorder='big')

        if 'argument' in instruction:

            bytecode += util.safe_decode(instruction['argument'])

    return bytecode



DEFAULT_FUNCTION_VISIBILITY = '100'
INTEGER_OVERFLOW_AND_UNDERFLOW = '101'
OUTDATED_COMPILER_VERSION = '102'
FLOATING_PRAGMA = '103'
UNCHECKED_RET_VAL = '104'
UNPROTECTED_ETHER_WITHDRAWAL = '105'
UNPROTECTED_SELFDESTRUCT = '106'
REENTRANCY = '107'
DEFAULT_STATE_VARIABLE_VISIBILITY = '108'
UNINITIALIZED_STORAGE_POINTER = '109'
ASSERT_VIOLATION = '110'
DEPRICATED_FUNCTIONS_USAGE = '111'
DELEGATECALL_TO_UNTRUSTED_CONTRACT = '112'
MULTIPLE_SENDS = '113'
TX_ORDER_DEPENDENCE = '114'
TX_ORIGIN_USAGE = '115'
TIMESTAMP_DEPENDENCE = '116'
# TODO: SWC ID 116 is missing, Add it if it's added to the https://github.com/SmartContractSecurity/SWC-registry
INCORRECT_CONSTRUCTOR_NAME = '118'
SHADOWING_STATE_VARIABLES = '119'
WEAK_RANDOMNESS = '120'
SIGNATURE_REPLAY = '121'
IMPROPER_VERIFICATION_BASED_ON_MSG_SENDER = '122'

PREDICTABLE_VARS_DEPENDENCE = 'N/A'    # TODO: Add the swc id when this is added to the SWC Registry



from mythril.laser.ethereum import svm
from mythril.laser.ethereum.state import Account
from mythril.ether.soliditycontract import SolidityContract
import copy
import logging
from .ops import get_variable, SStore, Call, VarType
from mythril.laser.ethereum.strategy.basic import DepthFirstSearchStrategy, BreadthFirstSearchStrategy, \
    ReturnRandomNaivelyStrategy, ReturnWeightedRandomStrategy


class SymExecWrapper:

    """
    Wrapper class for the LASER Symbolic virtual machine. Symbolically executes the code and does a bit of pre-analysis for convenience.
    """

    def __init__(self, contract, address, strategy, dynloader=None, max_depth=22,
                 execution_timeout=None, create_timeout=None, max_transaction_count=3):

        if strategy == 'dfs':
            s_strategy = DepthFirstSearchStrategy
        elif strategy == 'bfs':
            s_strategy = BreadthFirstSearchStrategy
        elif strategy == 'naive-random':
            s_strategy = ReturnRandomNaivelyStrategy
        elif strategy == 'weighted-random':
            s_strategy = ReturnWeightedRandomStrategy
        else:
            raise ValueError("Invalid strategy argument supplied")

        account = Account(address, contract.disassembly, dynamic_loader=dynloader, contract_name=contract.name)

        self.accounts = {address: account}

        self.laser = svm.LaserEVM(self.accounts, dynamic_loader=dynloader, max_depth=max_depth,
                                  execution_timeout=execution_timeout, strategy=s_strategy,
                                  create_timeout=create_timeout,
                                  max_transaction_count=max_transaction_count)

        if isinstance(contract, SolidityContract):
            self.laser.sym_exec(creation_code=contract.creation_code, contract_name=contract.name)
        else:
            self.laser.sym_exec(address)

        self.nodes = self.laser.nodes
        self.edges = self.laser.edges

        # Generate lists of interesting operations

        self.calls = []
        self.sstors = {}

        for key in self.nodes:

            state_index = 0

            for state in self.nodes[key].states:

                instruction = state.get_current_instruction()

                op = instruction['opcode']

                if op in ('CALL', 'CALLCODE', 'DELEGATECALL', 'STATICCALL'):

                    stack = state.mstate.stack

                    if op in ('CALL', 'CALLCODE'):
                        gas, to, value, meminstart, meminsz, memoutstart, memoutsz = \
                            get_variable(stack[-1]), get_variable(stack[-2]), get_variable(stack[-3]), get_variable(stack[-4]), get_variable(stack[-5]), get_variable(stack[-6]), get_variable(stack[-7])

                        if to.type == VarType.CONCRETE and to.val < 5:
                                # ignore prebuilts
                                continue

                        if meminstart.type == VarType.CONCRETE and meminsz.type == VarType.CONCRETE:
                            self.calls.append(Call(self.nodes[key], state, state_index, op, to, gas, value, state.mstate.memory[meminstart.val:meminsz.val * 4]))
                        else:
                            self.calls.append(Call(self.nodes[key], state, state_index, op, to, gas, value))
                    else:
                        gas, to, meminstart, meminsz, memoutstart, memoutsz = \
                            get_variable(stack[-1]), get_variable(stack[-2]), get_variable(stack[-3]), get_variable(stack[-4]), get_variable(stack[-5]), get_variable(stack[-6])

                        self.calls.append(Call(self.nodes[key], state, state_index, op, to, gas))

                elif op == 'SSTORE':
                    stack = copy.deepcopy(state.mstate.stack)
                    address = state.environment.active_account.address

                    index, value = stack.pop(), stack.pop()

                    try:
                        self.sstors[address]
                    except KeyError:
                        self.sstors[address] = {}

                    try:
                        self.sstors[address][str(index)].append(SStore(self.nodes[key], state, state_index, value))
                    except KeyError:
                        self.sstors[address][str(index)] = [SStore(self.nodes[key], state, state_index, value)]

                state_index += 1

    def find_storage_write(self, address, index):

        # Find an SSTOR not constrained by caller that writes to storage index "index"

        try:
            for s in self.sstors[address][index]:

                taint = True

                for constraint in s.node.constraints:
                    if "caller" in str(constraint):
                        taint = False
                        break

                if taint:
                    return s.node.function_name

            return None
        except KeyError:
            return None



import binascii
import logging
from copy import copy, deepcopy

from ethereum import utils
from z3 import Extract, UDiv, simplify, Concat, ULT, UGT, BitVecNumRef, Not, \
    is_false, is_expr, ExprRef, URem, SRem, BitVec, Solver, is_true, BitVecVal, If, BoolRef, Or

import mythril.laser.ethereum.natives as natives
import mythril.laser.ethereum.util as helper
from mythril.laser.ethereum import util
from mythril.laser.ethereum.call import get_call_parameters
from mythril.laser.ethereum.evm_exceptions import VmException, StackUnderflowException, InvalidJumpDestination, \
    InvalidInstruction
from mythril.laser.ethereum.keccak import KeccakFunctionManager
from mythril.laser.ethereum.state import GlobalState, CalldataType
from mythril.laser.ethereum.transaction import MessageCallTransaction, TransactionStartSignal, \
    ContractCreationTransaction

TT256 = 2 ** 256
TT256M1 = 2 ** 256 - 1

keccak_function_manager = KeccakFunctionManager()


class StateTransition(object):
    """Decorator that handles global state copy and original return.

    This decorator calls the decorated instruction mutator function on a copy of the state that
    is passed to it. After the call, the resulting new states' program counter is automatically
    incremented if `increment_pc=True`.
    """

    def __init__(self, increment_pc=True):
        self.increment_pc = increment_pc

    @staticmethod
    def call_on_state_copy(func, func_obj, state):
        global_state_copy = copy(state)
        return func(func_obj, global_state_copy)

    def increment_states_pc(self, states):
        if self.increment_pc:
            for state in states:
                state.mstate.pc += 1
        return states

    def __call__(self, func):
        def wrapper(func_obj, global_state):
            new_global_states = self.call_on_state_copy(
                func,
                func_obj,
                global_state
            )
            return self.increment_states_pc(new_global_states)
        return wrapper


class Instruction:
    """
    Instruction class is used to mutate a state according to the current instruction
    """

    def __init__(self, op_code, dynamic_loader):
        self.dynamic_loader = dynamic_loader
        self.op_code = op_code

    def evaluate(self, global_state, post=False):
        """ Performs the mutation for this instruction """
        # Generalize some ops
        logging.debug("Evaluating {}".format(self.op_code))
        op = self.op_code.lower()
        if self.op_code.startswith("PUSH"):
            op = "push"
        elif self.op_code.startswith("DUP"):
            op = "dup"
        elif self.op_code.startswith("SWAP"):
            op = "swap"
        elif self.op_code.startswith("LOG"):
            op = "log"

        instruction_mutator = getattr(self, op + '_', None) if not post else getattr(self, op + '_' + 'post', None)

        if instruction_mutator is None:
            raise NotImplementedError

        return instruction_mutator(global_state)

    @StateTransition()
    def jumpdest_(self, global_state):
        return [global_state]

    @StateTransition()
    def push_(self, global_state):
        push_instruction = global_state.get_current_instruction()
        push_value = push_instruction['argument'][2:]

        try:
            length_of_value = 2*int(push_instruction['opcode'][4:])
        except ValueError:
            raise VmException('Invalid Push instruction')

        push_value += "0" * max(length_of_value - len(push_value), 0)
        global_state.mstate.stack.append(BitVecVal(int(push_value, 16), 256))
        return [global_state]

    @StateTransition()
    def dup_(self, global_state):
        value = int(global_state.get_current_instruction()['opcode'][3:], 10)
        global_state.mstate.stack.append(global_state.mstate.stack[-value])
        return [global_state]

    @StateTransition()
    def swap_(self, global_state):
        depth = int(self.op_code[4:])
        stack = global_state.mstate.stack
        stack[-depth - 1], stack[-1] = stack[-1], stack[-depth - 1]
        return [global_state]

    @StateTransition()
    def pop_(self, global_state):
        global_state.mstate.stack.pop()
        return [global_state]

    @StateTransition()
    def and_(self, global_state):
        stack = global_state.mstate.stack
        op1, op2 = stack.pop(), stack.pop()
        if type(op1) == BoolRef:
            op1 = If(op1, BitVecVal(1, 256), BitVecVal(0, 256))
        if type(op2) == BoolRef:
            op2 = If(op2, BitVecVal(1, 256), BitVecVal(0, 256))

        stack.append(op1 & op2)

        return [global_state]

    @StateTransition()
    def or_(self, global_state):
        stack = global_state.mstate.stack
        op1, op2 = stack.pop(), stack.pop()

        if type(op1) == BoolRef:
            op1 = If(op1, BitVecVal(1, 256), BitVecVal(0, 256))

        if type(op2) == BoolRef:
            op2 = If(op2, BitVecVal(1, 256), BitVecVal(0, 256))

        stack.append(op1 | op2)

        return [global_state]

    @StateTransition()
    def xor_(self, global_state):
        mstate = global_state.mstate
        mstate.stack.append(mstate.stack.pop() ^ mstate.stack.pop())
        return [global_state]

    @StateTransition()
    def not_(self, global_state: GlobalState):
        mstate = global_state.mstate
        mstate.stack.append(TT256M1 - mstate.stack.pop())
        return [global_state]

    @StateTransition()
    def byte_(self, global_state):
        mstate = global_state.mstate
        op0, op1 = mstate.stack.pop(), mstate.stack.pop()
        if not isinstance(op1, ExprRef):
            op1 = BitVecVal(op1, 256)
        try:
            index = util.get_concrete_int(op0)
            offset = (31 - index) * 8
            if offset >= 0:
                result = simplify(Concat(BitVecVal(0, 248), Extract(offset + 7, offset, op1)))
            else:
                result = 0
        except TypeError:
            logging.debug("BYTE: Unsupported symbolic byte offset")
            result = global_state.new_bitvec(str(simplify(op1)) + "[" + str(simplify(op0)) + "]", 256)

        mstate.stack.append(result)
        return [global_state]

    # Arithmetic
    @StateTransition()
    def add_(self, global_state):
        global_state.mstate.stack.append(
            (helper.pop_bitvec(global_state.mstate) + helper.pop_bitvec(global_state.mstate)))
        return [global_state]

    @StateTransition()
    def sub_(self, global_state):
        global_state.mstate.stack.append(
            (helper.pop_bitvec(global_state.mstate) - helper.pop_bitvec(global_state.mstate)))
        return [global_state]

    @StateTransition()
    def mul_(self, global_state):
        global_state.mstate.stack.append(
            (helper.pop_bitvec(global_state.mstate) * helper.pop_bitvec(global_state.mstate)))
        return [global_state]

    @StateTransition()
    def div_(self, global_state):
        op0, op1 = util.pop_bitvec(global_state.mstate), util.pop_bitvec(global_state.mstate)
        if op1 == 0:
            global_state.mstate.stack.append(BitVecVal(0, 256))
        else:
            global_state.mstate.stack.append(UDiv(op0, op1))
        return [global_state]

    @StateTransition()
    def sdiv_(self, global_state):
        s0, s1 = util.pop_bitvec(global_state.mstate), util.pop_bitvec(global_state.mstate)
        if s1 == 0:
            global_state.mstate.stack.append(BitVecVal(0, 256))
        else:
            global_state.mstate.stack.append(s0 / s1)
        return [global_state]

    @StateTransition()
    def mod_(self, global_state):
        s0, s1 = util.pop_bitvec(global_state.mstate), util.pop_bitvec(global_state.mstate)
        global_state.mstate.stack.append(0 if s1 == 0 else URem(s0, s1))
        return [global_state]

    @StateTransition()
    def smod_(self, global_state):
        s0, s1 = util.pop_bitvec(global_state.mstate), util.pop_bitvec(global_state.mstate)
        global_state.mstate.stack.append(0 if s1 == 0 else SRem(s0, s1))
        return [global_state]

    @StateTransition()
    def addmod_(self, global_state):
        s0, s1, s2 = util.pop_bitvec(global_state.mstate), util.pop_bitvec(global_state.mstate), util.pop_bitvec(
            global_state.mstate)
        global_state.mstate.stack.append(URem(URem(s0, s2) + URem(s1, s2), s2))
        return [global_state]

    @StateTransition()
    def mulmod_(self, global_state):
        s0, s1, s2 = util.pop_bitvec(global_state.mstate), util.pop_bitvec(global_state.mstate), util.pop_bitvec(
            global_state.mstate)
        global_state.mstate.stack.append(URem(URem(s0, s2) * URem(s1, s2), s2))
        return [global_state]

    @StateTransition()
    def exp_(self, global_state):
        state = global_state.mstate

        base, exponent = util.pop_bitvec(state), util.pop_bitvec(state)
        if (type(base) != BitVecNumRef) or (type(exponent) != BitVecNumRef):
            state.stack.append(global_state.new_bitvec("(" + str(simplify(base)) + ")**(" + str(simplify(exponent)) + ")", 256))
        else:
            state.stack.append(pow(base.as_long(), exponent.as_long(), 2**256))

        return [global_state]

    @StateTransition()
    def signextend_(self, global_state):
        state = global_state.mstate
        s0, s1 = state.stack.pop(), state.stack.pop()

        try:
            s0 = util.get_concrete_int(s0)
            s1 = util.get_concrete_int(s1)
        except TypeError:
            return []

        if s0 <= 31:
            testbit = s0 * 8 + 7
            if s1 & (1 << testbit):
                state.stack.append(s1 | (TT256 - (1 << testbit)))
            else:
                state.stack.append(s1 & ((1 << testbit) - 1))
        else:
            state.stack.append(s1)

        return [global_state]

    # Comparisons
    @StateTransition()
    def lt_(self, global_state):
        state = global_state.mstate
        exp = ULT(util.pop_bitvec(state), util.pop_bitvec(state))
        state.stack.append(exp)
        return [global_state]

    @StateTransition()
    def gt_(self, global_state):
        state = global_state.mstate
        exp = UGT(util.pop_bitvec(state), util.pop_bitvec(state))
        state.stack.append(exp)
        return [global_state]

    @StateTransition()
    def slt_(self, global_state):
        state = global_state.mstate
        exp = util.pop_bitvec(state) < util.pop_bitvec(state)
        state.stack.append(exp)
        return [global_state]

    @StateTransition()
    def sgt_(self, global_state):
        state = global_state.mstate

        exp = util.pop_bitvec(state) > util.pop_bitvec(state)
        state.stack.append(exp)
        return [global_state]

    @StateTransition()
    def eq_(self, global_state):
        state = global_state.mstate

        op1 = state.stack.pop()
        op2 = state.stack.pop()

        if type(op1) == BoolRef:
            op1 = If(op1, BitVecVal(1, 256), BitVecVal(0, 256))

        if type(op2) == BoolRef:
            op2 = If(op2, BitVecVal(1, 256), BitVecVal(0, 256))

        exp = op1 == op2

        state.stack.append(exp)
        return [global_state]

    @StateTransition()
    def iszero_(self, global_state):
        state = global_state.mstate

        val = state.stack.pop()
        exp = val == False if type(val) == BoolRef else val == 0
        state.stack.append(exp)

        return [global_state]

    # Call data
    @StateTransition()
    def callvalue_(self, global_state):
        state = global_state.mstate
        environment = global_state.environment
        state.stack.append(environment.callvalue)

        return [global_state]

    @StateTransition()
    def calldataload_(self, global_state):
        state = global_state.mstate
        environment = global_state.environment
        op0 = state.stack.pop()

        try:
            offset = util.get_concrete_int(simplify(op0))
            b = environment.calldata[offset]
        except TypeError:
            logging.debug("CALLDATALOAD: Unsupported symbolic index")
            state.stack.append(global_state.new_bitvec(
                "calldata_" + str(environment.active_account.contract_name) + "[" + str(simplify(op0)) + "]", 256))
            return [global_state]
        except IndexError:
            logging.debug("Calldata not set, using symbolic variable instead")
            state.stack.append(global_state.new_bitvec(
                "calldata_" + str(environment.active_account.contract_name) + "[" + str(simplify(op0)) + "]", 256))
            return [global_state]

        if type(b) == int:

            try:
                val = b''.join([calldata.to_bytes(1, byteorder='big') for calldata in
                                environment.calldata[offset:offset+32]])

                logging.debug("Final value: " + str(int.from_bytes(val, byteorder='big')))
                state.stack.append(BitVecVal(int.from_bytes(val, byteorder='big'), 256))

            except (TypeError, AttributeError):
                state.stack.append(global_state.new_bitvec(
                    "calldata_" + str(environment.active_account.contract_name) + "[" + str(simplify(op0)) + "]", 256))
        else:
            # symbolic variable
            state.stack.append(global_state.new_bitvec(
                "calldata_" + str(environment.active_account.contract_name) + "[" + str(simplify(op0)) + "]", 256))

        return [global_state]

    @StateTransition()
    def calldatasize_(self, global_state):
        state = global_state.mstate
        environment = global_state.environment
        if environment.calldata_type == CalldataType.SYMBOLIC:
            state.stack.append(global_state.new_bitvec("calldatasize_" + environment.active_account.contract_name, 256))
        else:
            state.stack.append(BitVecVal(len(environment.calldata), 256))
        return [global_state]

    @StateTransition()
    def calldatacopy_(self, global_state):
        state = global_state.mstate
        environment = global_state.environment
        op0, op1, op2 = state.stack.pop(), state.stack.pop(), state.stack.pop()

        try:
            mstart = util.get_concrete_int(op0)
        except TypeError:
            logging.debug("Unsupported symbolic memory offset in CALLDATACOPY")
            return [global_state]

        dstart_sym = False
        try:
            dstart = util.get_concrete_int(op1)
        except TypeError:
            logging.debug("Unsupported symbolic calldata offset in CALLDATACOPY")
            dstart = simplify(op1)
            dstart_sym = True

        size_sym = False
        try:
            size = util.get_concrete_int(op2)
        except TypeError:
            logging.debug("Unsupported symbolic size in CALLDATACOPY")
            size = simplify(op2)
            size_sym = True

        if dstart_sym or size_sym:
            state.mem_extend(mstart, 1)
            state.memory[mstart] = global_state.new_bitvec(
                "calldata_" + str(environment.active_account.contract_name) + "[" + str(dstart) + ": + " + str(
                    size) + "]", 256)
            return [global_state]

        if size > 0:
            try:
                state.mem_extend(mstart, size)
            except TypeError:
                logging.debug("Memory allocation error: mstart = " + str(mstart) + ", size = " + str(size))
                state.mem_extend(mstart, 1)
                state.memory[mstart] = global_state.new_bitvec(
                    "calldata_" + str(environment.active_account.contract_name) + "[" + str(dstart) + ": + " + str(
                        size) + "]", 256)
                return [global_state]

            try:
                i_data = environment.calldata[dstart]

                for i in range(mstart, mstart + size):
                    state.memory[i] = environment.calldata[i_data]
                    i_data += 1
            except IndexError:
                logging.debug("Exception copying calldata to memory")

                state.memory[mstart] = global_state.new_bitvec(
                    "calldata_" + str(environment.active_account.contract_name) + "[" + str(dstart) + ": + " + str(
                        size) + "]", 256)
        return [global_state]

    # Environment
    @StateTransition()
    def address_(self, global_state):
        state = global_state.mstate
        environment = global_state.environment
        state.stack.append(environment.address)
        return [global_state]

    @StateTransition()
    def balance_(self, global_state):
        state = global_state.mstate
        address = state.stack.pop()
        state.stack.append(global_state.new_bitvec("balance_at_" + str(address), 256))
        return [global_state]

    @StateTransition()
    def origin_(self, global_state):
        state = global_state.mstate
        environment = global_state.environment
        state.stack.append(environment.origin)
        return [global_state]

    @StateTransition()
    def caller_(self, global_state):
        state = global_state.mstate
        environment = global_state.environment
        state.stack.append(environment.sender)
        return [global_state]

    @StateTransition()
    def codesize_(self, global_state):
        state = global_state.mstate
        environment = global_state.environment
        disassembly = environment.code
        state.stack.append(len(disassembly.bytecode) // 2)
        return [global_state]

    @StateTransition()
    def sha3_(self, global_state):
        global keccak_function_manager

        state = global_state.mstate
        op0, op1 = state.stack.pop(), state.stack.pop()

        try:
            index, length = util.get_concrete_int(op0), util.get_concrete_int(op1)
        except TypeError:
            # Can't access symbolic memory offsets
            if is_expr(op0):
                op0 = simplify(op0)
            state.stack.append(BitVec("KECCAC_mem[" + str(op0) + "]", 256))
            return [global_state]

        try:
            state.mem_extend(index, length)
            data = b''.join([util.get_concrete_int(i).to_bytes(1, byteorder='big')
                             for i in state.memory[index: index + length]])

        except TypeError:
            argument = str(state.memory[index]).replace(" ", "_")

            result = BitVec("KECCAC[{}]".format(argument), 256)
            keccak_function_manager.add_keccak(result, state.memory[index])
            state.stack.append(result)
            return [global_state]

        keccak = utils.sha3(utils.bytearray_to_bytestr(data))
        logging.debug("Computed SHA3 Hash: " + str(binascii.hexlify(keccak)))

        state.stack.append(BitVecVal(util.concrete_int_from_bytes(keccak, 0), 256))
        return [global_state]

    @StateTransition()
    def gasprice_(self, global_state):
        global_state.mstate.stack.append(global_state.new_bitvec("gasprice", 256))
        return [global_state]

    @StateTransition()
    def codecopy_(self, global_state):
        memory_offset, code_offset, size = global_state.mstate.stack.pop(), global_state.mstate.stack.pop(), global_state.mstate.stack.pop()

        try:
            concrete_memory_offset = helper.get_concrete_int(memory_offset)
        except TypeError:
            logging.debug("Unsupported symbolic memory offset in CODECOPY")
            return [global_state]

        try:
            concrete_size = helper.get_concrete_int(size)
            global_state.mstate.mem_extend(concrete_memory_offset, concrete_size)
        except TypeError:
            # except both attribute error and Exception
            global_state.mstate.mem_extend(concrete_memory_offset, 1)
            global_state.mstate.memory[concrete_memory_offset] = \
                global_state.new_bitvec("code({})".format(global_state.environment.active_account.contract_name), 256)
            return [global_state]

        try:
            concrete_code_offset = helper.get_concrete_int(code_offset)
        except TypeError:
            logging.debug("Unsupported symbolic code offset in CODECOPY")
            global_state.mstate.mem_extend(concrete_memory_offset, concrete_size)
            for i in range(concrete_size):
                global_state.mstate.memory[concrete_memory_offset + i] = \
                    global_state.new_bitvec("code({})".format(global_state.environment.active_account.contract_name), 256)
            return [global_state]

        bytecode = global_state.environment.code.bytecode

        if concrete_size == 0 and isinstance(global_state.current_transaction, ContractCreationTransaction):
            if concrete_code_offset >= len(global_state.environment.code.bytecode) // 2:
                global_state.mstate.mem_extend(concrete_memory_offset, 1)
                global_state.mstate.memory[concrete_memory_offset] = \
                    global_state.new_bitvec("code({})".format(global_state.environment.active_account.contract_name), 256)
                return [global_state]

        for i in range(concrete_size):
            if 2 * (concrete_code_offset + i + 1) <= len(bytecode):
                global_state.mstate.memory[concrete_memory_offset + i] =\
                    int(bytecode[2*(concrete_code_offset + i): 2*(concrete_code_offset + i + 1)], 16)
            else:
                global_state.mstate.memory[concrete_memory_offset + i] = \
                    global_state.new_bitvec("code({})".format(global_state.environment.active_account.contract_name), 256)

        return [global_state]

    @StateTransition()
    def extcodesize_(self, global_state):
        state = global_state.mstate
        addr = state.stack.pop()
        environment = global_state.environment
        try:
            addr = hex(helper.get_concrete_int(addr))
        except TypeError:
            logging.info("unsupported symbolic address for EXTCODESIZE")
            state.stack.append(global_state.new_bitvec("extcodesize_" + str(addr), 256))
            return [global_state]

        try:
            code = self.dynamic_loader.dynld(environment.active_account.address, addr)
        except Exception as e:
            logging.info("error accessing contract storage due to: " + str(e))
            state.stack.append(global_state.new_bitvec("extcodesize_" + str(addr), 256))
            return [global_state]

        if code is None:
            state.stack.append(0)
        else:
            state.stack.append(len(code.bytecode) // 2)

        return [global_state]

    @StateTransition()
    def extcodecopy_(self, global_state):
        # FIXME: not implemented
        state = global_state.mstate
        addr = state.stack.pop()
        start, s2, size = state.stack.pop(), state.stack.pop(), state.stack.pop()
        return [global_state]

    @StateTransition()
    def returndatasize_(self, global_state):
        global_state.mstate.stack.append(global_state.new_bitvec("returndatasize", 256))
        return [global_state]

    @StateTransition()
    def blockhash_(self, global_state):
        state = global_state.mstate
        blocknumber = state.stack.pop()
        state.stack.append(global_state.new_bitvec("blockhash_block_" + str(blocknumber), 256))
        return [global_state]

    @StateTransition()
    def coinbase_(self, global_state):
        global_state.mstate.stack.append(global_state.new_bitvec("coinbase", 256))
        return [global_state]

    @StateTransition()
    def timestamp_(self, global_state):
        global_state.mstate.stack.append(global_state.new_bitvec("timestamp", 256))
        return [global_state]

    @StateTransition()
    def number_(self, global_state):
        global_state.mstate.stack.append(global_state.new_bitvec("block_number", 256))
        return [global_state]

    @StateTransition()
    def difficulty_(self, global_state):
        global_state.mstate.stack.append(global_state.new_bitvec("block_difficulty", 256))
        return [global_state]

    @StateTransition()
    def gaslimit_(self, global_state):
        global_state.mstate.stack.append(global_state.new_bitvec("block_gaslimit", 256))
        return [global_state]

    # Memory operations
    @StateTransition()
    def mload_(self, global_state):
        state = global_state.mstate
        op0 = state.stack.pop()

        logging.debug("MLOAD[" + str(op0) + "]")

        try:
            offset = util.get_concrete_int(op0)
        except TypeError:
            logging.debug("Can't MLOAD from symbolic index")
            data = global_state.new_bitvec("mem[" + str(simplify(op0)) + "]", 256)
            state.stack.append(data)
            return [global_state]

        try:
            data = util.concrete_int_from_bytes(state.memory, offset)
        except IndexError:  # Memory slot not allocated
            data = global_state.new_bitvec("mem[" + str(offset) + "]", 256)
        except TypeError:  # Symbolic memory
            data = state.memory[offset]

        logging.debug("Load from memory[" + str(offset) + "]: " + str(data))

        state.stack.append(data)
        return [global_state]

    @StateTransition()
    def mstore_(self, global_state):
        state = global_state.mstate
        op0, value = state.stack.pop(), state.stack.pop()

        try:
            mstart = util.get_concrete_int(op0)
        except TypeError:
            logging.debug("MSTORE to symbolic index. Not supported")
            return [global_state]

        try:
            state.mem_extend(mstart, 32)
        except Exception:
            logging.debug("Error extending memory, mstart = " + str(mstart) + ", size = 32")

        logging.debug("MSTORE to mem[" + str(mstart) + "]: " + str(value))

        try:
            # Attempt to concretize value
            _bytes = util.concrete_int_to_bytes(value)
            state.memory[mstart: mstart + len(_bytes)] = _bytes
        except:
            try:
                state.memory[mstart] = value
            except TypeError:
                logging.debug("Invalid memory access")

        return [global_state]

    @StateTransition()
    def mstore8_(self, global_state):
        state = global_state.mstate
        op0, value = state.stack.pop(), state.stack.pop()

        try:
            offset = util.get_concrete_int(op0)
        except TypeError:
            logging.debug("MSTORE to symbolic index. Not supported")
            return [global_state]

        state.mem_extend(offset, 1)

        state.memory[offset] = value % 256
        return [global_state]

    @StateTransition()
    def sload_(self, global_state):
        global keccak_function_manager
        
        state = global_state.mstate
        index = state.stack.pop()
        logging.debug("Storage access at index " + str(index))

        try:
            index = util.get_concrete_int(index)
            return self._sload_helper(global_state, index)

        except TypeError:
            if not keccak_function_manager.is_keccak(index):
                return self._sload_helper(global_state, str(index))

            storage_keys = global_state.environment.active_account.storage.keys()
            keccak_keys = list(filter(keccak_function_manager.is_keccak, storage_keys))

            results = []
            constraints = []

            for keccak_key in keccak_keys:
                key_argument = keccak_function_manager.get_argument(keccak_key)
                index_argument = keccak_function_manager.get_argument(index)
                constraints.append((keccak_key, key_argument == index_argument))

            for (keccak_key, constraint) in constraints:
                if constraint in state.constraints:
                    results += self._sload_helper(global_state, keccak_key, [constraint])
            if len(results) > 0:
                return results

            for (keccak_key, constraint) in constraints:
                results += self._sload_helper(copy(global_state), keccak_key, [constraint])
            if len(results) > 0:
                return results
            
            return self._sload_helper(global_state, str(index))

    @staticmethod
    def _sload_helper(global_state, index, constraints=None):
        try:
            data = global_state.environment.active_account.storage[index]
        except KeyError:
            data = global_state.new_bitvec("storage_" + str(index), 256)
            global_state.environment.active_account.storage[index] = data

        if constraints is not None:
            global_state.mstate.constraints += constraints

        global_state.mstate.stack.append(data)
        return [global_state]

    @staticmethod
    def _get_constraints(keccak_keys, this_key, argument):
        global keccak_function_manager
        for keccak_key in keccak_keys:
            if keccak_key == this_key:
                continue
            keccak_argument = keccak_function_manager.get_argument(keccak_key)
            yield keccak_argument != argument

    @StateTransition()
    def sstore_(self, global_state):
        global keccak_function_manager
        state = global_state.mstate
        index, value = state.stack.pop(), state.stack.pop()

        logging.debug("Write to storage[" + str(index) + "]")

        try:
            index = util.get_concrete_int(index)
            return self._sstore_helper(global_state, index, value)
        except TypeError:
            is_keccak = keccak_function_manager.is_keccak(index)
            if not is_keccak:
                return self._sstore_helper(global_state, str(index), value)

            storage_keys = global_state.environment.active_account.storage.keys()
            keccak_keys = filter(keccak_function_manager.is_keccak, storage_keys)

            solver = Solver()
            solver.set(timeout=1000)

            results = []
            new = False

            for keccak_key in keccak_keys:
                key_argument = keccak_function_manager.get_argument(keccak_key)
                index_argument = keccak_function_manager.get_argument(index)

                if is_true(key_argument == index_argument):
                    return self._sstore_helper(copy(global_state), keccak_key, value, key_argument == index_argument)

                results += self._sstore_helper(copy(global_state), keccak_key, value, key_argument == index_argument)

                new = Or(new, key_argument != index_argument)

            if len(results) > 0:
                results += self._sstore_helper(copy(global_state), str(index), value, new)
                return results

            return self._sstore_helper(global_state, str(index), value)

    @staticmethod
    def _sstore_helper(global_state, index, value, constraint=None):
        try:
            global_state.environment.active_account = deepcopy(global_state.environment.active_account)
            global_state.accounts[
                global_state.environment.active_account.address] = global_state.environment.active_account

            global_state.environment.active_account.storage[index] =\
                value if not isinstance(value, ExprRef) else simplify(value)
        except KeyError:
            logging.debug("Error writing to storage: Invalid index")

        if constraint is not None:
            global_state.mstate.constraints.append(constraint)

        return [global_state]

    @StateTransition(increment_pc=False)
    def jump_(self, global_state):
        state = global_state.mstate
        disassembly = global_state.environment.code
        try:
            jump_addr = util.get_concrete_int(state.stack.pop())
        except TypeError:
            raise InvalidJumpDestination("Invalid jump argument (symbolic address)")
        except IndexError:
            raise StackUnderflowException()

        index = util.get_instruction_index(disassembly.instruction_list, jump_addr)
        if index is None:
            raise InvalidJumpDestination("JUMP to invalid address")

        op_code = disassembly.instruction_list[index]['opcode']

        if op_code != "JUMPDEST":
            raise InvalidJumpDestination("Skipping JUMP to invalid destination (not JUMPDEST): " + str(jump_addr))

        new_state = copy(global_state)
        new_state.mstate.pc = index
        new_state.mstate.depth += 1

        return [new_state]

    @StateTransition(increment_pc=False)
    def jumpi_(self, global_state):
        state = global_state.mstate
        disassembly = global_state.environment.code
        states = []

        op0, condition = state.stack.pop(), state.stack.pop()

        try:
            jump_addr = util.get_concrete_int(op0)
        except TypeError:
            logging.debug("Skipping JUMPI to invalid destination.")
            global_state.mstate.pc += 1
            return [global_state]

        # False case
        negated = simplify(Not(condition)) if type(condition) == BoolRef else condition == 0

        if (type(negated) == bool and negated) or (type(negated) == BoolRef and not is_false(negated)):
            new_state = copy(global_state)
            new_state.mstate.depth += 1
            new_state.mstate.pc += 1
            new_state.mstate.constraints.append(negated)
            states.append(new_state)
        else:
            logging.debug("Pruned unreachable states.")

        # True case

        # Get jump destination
        index = util.get_instruction_index(disassembly.instruction_list, jump_addr)
        if not index:
            logging.debug("Invalid jump destination: " + str(jump_addr))
            return states

        instr = disassembly.instruction_list[index]

        condi = simplify(condition) if type(condition) == BoolRef else condition != 0
        if instr['opcode'] == "JUMPDEST":
            if (type(condi) == bool and condi) or (type(condi) == BoolRef and not is_false(condi)):
                new_state = copy(global_state)
                new_state.mstate.pc = index
                new_state.mstate.depth += 1
                new_state.mstate.constraints.append(condi)
                states.append(new_state)
            else:
                logging.debug("Pruned unreachable states.")
        return states

    @StateTransition()
    def pc_(self, global_state):
        global_state.mstate.stack.append(global_state.mstate.pc - 1)
        return [global_state]

    @StateTransition()
    def msize_(self, global_state):
        global_state.mstate.stack.append(global_state.new_bitvec("msize", 256))
        return [global_state]

    @StateTransition()
    def gas_(self, global_state):
        global_state.mstate.stack.append(global_state.new_bitvec("gas", 256))
        return [global_state]

    @StateTransition()
    def log_(self, global_state):
        # TODO: implement me
        state = global_state.mstate
        dpth = int(self.op_code[3:])
        state.stack.pop(), state.stack.pop()
        [state.stack.pop() for _ in range(dpth)]
        # Not supported
        return [global_state]

    @StateTransition()
    def create_(self, global_state):
        # TODO: implement me
        state = global_state.mstate
        state.stack.pop(), state.stack.pop(), state.stack.pop()
        # Not supported
        state.stack.append(0)
        return [global_state]

    @StateTransition()
    def return_(self, global_state):
        state = global_state.mstate
        offset, length = state.stack.pop(), state.stack.pop()
        return_data = [global_state.new_bitvec("return_data", 256)]
        try:
            return_data = state.memory[util.get_concrete_int(offset):util.get_concrete_int(offset + length)]
        except TypeError:
            logging.debug("Return with symbolic length or offset. Not supported")
        global_state.current_transaction.end(global_state, return_data)

    @StateTransition()
    def suicide_(self, global_state):
        target = global_state.mstate.stack.pop()

        # Often the target of the suicide instruction will be symbolic
        # If it isn't then well transfer the balance to the indicated contract
        if isinstance(target, BitVecNumRef):
            target = '0x' + hex(target.as_long())[-40:]
        if isinstance(target, str):
            try:
                global_state.world_state[target].balance += global_state.environment.active_account.balance
            except KeyError:
                global_state.world_state.create_account(address=target, balance=global_state.environment.active_account.balance)

        global_state.environment.active_account.balance = 0
        global_state.environment.active_account.deleted = True

        global_state.current_transaction.end(global_state)

    @StateTransition()
    def revert_(self, global_state):
        state = global_state.mstate
        offset, length = state.stack.pop(), state.stack.pop()
        return_data = [global_state.new_bitvec("return_data", 256)]
        try:
            return_data = state.memory[util.get_concrete_int(offset):util.get_concrete_int(offset + length)]
        except TypeError:
            logging.debug("Return with symbolic length or offset. Not supported")
        global_state.current_transaction.end(global_state, return_data=return_data, revert=True)

    @StateTransition()
    def assert_fail_(self, global_state):
        # 0xfe: designated invalid opcode
        raise InvalidInstruction

    @StateTransition()
    def invalid_(self, global_state):
        raise InvalidInstruction

    @StateTransition()
    def stop_(self, global_state):
        global_state.current_transaction.end(global_state)

    @StateTransition()
    def call_(self, global_state):

        instr = global_state.get_current_instruction()
        environment = global_state.environment

        try:
            callee_address, callee_account, call_data, value, call_data_type, gas, memory_out_offset, memory_out_size = get_call_parameters(
                global_state, self.dynamic_loader, True)
        except ValueError as e:
            logging.debug(
                "Could not determine required parameters for call, putting fresh symbol on the stack. \n{}".format(e)
            )
            # TODO: decide what to do in this case
            global_state.mstate.stack.append(global_state.new_bitvec("retval_" + str(instr['address']), 256))
            return [global_state]
        global_state.mstate.stack.append(global_state.new_bitvec("retval_" + str(instr['address']), 256))

        if 0 < int(callee_address, 16) < 5:
            logging.info("Native contract called: " + callee_address)
            if call_data == [] and call_data_type == CalldataType.SYMBOLIC:
                logging.debug("CALL with symbolic data not supported")
                return [global_state]

            try:
                mem_out_start = helper.get_concrete_int(memory_out_offset)
                mem_out_sz = memory_out_size.as_long()
            except TypeError:
                logging.debug("CALL with symbolic start or offset not supported")
                return [global_state]

            global_state.mstate.mem_extend(mem_out_start, mem_out_sz)
            call_address_int = int(callee_address, 16)
            try:
                data = natives.native_contracts(call_address_int, call_data)
            except natives.NativeContractException:
                contract_list = ['ecerecover', 'sha256', 'ripemd160', 'identity']
                for i in range(mem_out_sz):
                    global_state.mstate.memory[mem_out_start + i] = global_state.new_bitvec(contract_list[call_address_int - 1] +
                                                                           "(" + str(call_data) + ")", 256)

                return [global_state]

            for i in range(min(len(data), mem_out_sz)):  # If more data is used then it's chopped off
                global_state.mstate.memory[mem_out_start + i] = data[i]

            # TODO: maybe use BitVec here constrained to 1
            return [global_state]

        transaction = MessageCallTransaction(global_state.world_state,
                                             callee_account,
                                             BitVecVal(int(environment.active_account.address, 16), 256),
                                             call_data=call_data,
                                             gas_price=environment.gasprice,
                                             call_value=value,
                                             origin=environment.origin,
                                             call_data_type=call_data_type)
        raise TransactionStartSignal(transaction, self.op_code)

    @StateTransition()
    def call_post(self, global_state):
        instr = global_state.get_current_instruction()

        try:
            _, _, _, _, _, _, memory_out_offset, memory_out_size = get_call_parameters(
                global_state, self.dynamic_loader, True)
        except ValueError as e:
            logging.info(
                "Could not determine required parameters for call, putting fresh symbol on the stack. \n{}".format(e)
            )
            global_state.mstate.stack.append(global_state.new_bitvec("retval_" + str(instr['address']), 256))
            return [global_state]

        if global_state.last_return_data is None:
            # Put return value on stack
            return_value = global_state.new_bitvec("retval_" + str(instr['address']), 256)
            global_state.mstate.stack.append(return_value)
            global_state.mstate.constraints.append(return_value == 0)

            return [global_state]

        try:
            memory_out_offset = util.get_concrete_int(memory_out_offset) if isinstance(memory_out_offset, ExprRef) else memory_out_offset
            memory_out_size = util.get_concrete_int(memory_out_size) if isinstance(memory_out_size, ExprRef) else memory_out_size
        except TypeError:
            global_state.mstate.stack.append(global_state.new_bitvec("retval_" + str(instr['address']), 256))
            return [global_state]

        # Copy memory
        global_state.mstate.mem_extend(memory_out_offset, min(memory_out_size, len(global_state.last_return_data)))
        for i in range(min(memory_out_size, len(global_state.last_return_data))):
            global_state.mstate.memory[i + memory_out_offset] = global_state.last_return_data[i]

        # Put return value on stack
        return_value = global_state.new_bitvec("retval_" + str(instr['address']), 256)
        global_state.mstate.stack.append(return_value)
        global_state.mstate.constraints.append(return_value == 1)

        return [global_state]

    @StateTransition()
    def callcode_(self, global_state):
        instr = global_state.get_current_instruction()
        environment = global_state.environment

        try:
            callee_address, callee_account, call_data, value, call_data_type, gas, _, _ = get_call_parameters(
                global_state, self.dynamic_loader, True)
        except ValueError as e:
            logging.info(
                "Could not determine required parameters for call, putting fresh symbol on the stack. \n{}".format(e)
            )
            global_state.mstate.stack.append(global_state.new_bitvec("retval_" + str(instr['address']), 256))
            return [global_state]

        transaction = MessageCallTransaction(global_state.world_state,
                                             environment.active_account,
                                             environment.address,
                                             call_data=call_data,
                                             gas_price=environment.gasprice,
                                             call_value=value,
                                             origin=environment.origin,
                                             call_data_type=call_data_type,
                                             code=callee_account.code
                                             )
        raise TransactionStartSignal(transaction, self.op_code)

    @StateTransition()
    def callcode_post(self, global_state):
        instr = global_state.get_current_instruction()

        try:
            _, _, _, _, _, _, memory_out_offset, memory_out_size = get_call_parameters(
                global_state, self.dynamic_loader, True)
        except ValueError as e:
            logging.info(
                "Could not determine required parameters for call, putting fresh symbol on the stack. \n{}".format(e)
            )
            global_state.mstate.stack.append(global_state.new_bitvec("retval_" + str(instr['address']), 256))
            return [global_state]

        if global_state.last_return_data is None:
            # Put return value on stack
            return_value = global_state.new_bitvec("retval_" + str(instr['address']), 256)
            global_state.mstate.stack.append(return_value)
            global_state.mstate.constraints.append(return_value == 0)

            return [global_state]

        try:
            memory_out_offset = util.get_concrete_int(memory_out_offset) if isinstance(memory_out_offset, ExprRef) else memory_out_offset
            memory_out_size = util.get_concrete_int(memory_out_size) if isinstance(memory_out_size, ExprRef) else memory_out_size
        except TypeError:
            global_state.mstate.stack.append(global_state.new_bitvec("retval_" + str(instr['address']), 256))
            return [global_state]

        # Copy memory
        global_state.mstate.mem_extend(memory_out_offset, min(memory_out_size, len(global_state.last_return_data)))
        for i in range(min(memory_out_size, len(global_state.last_return_data))):
            global_state.mstate.memory[i + memory_out_offset] = global_state.last_return_data[i]

        # Put return value on stack
        return_value = global_state.new_bitvec("retval_" + str(instr['address']), 256)
        global_state.mstate.stack.append(return_value)
        global_state.mstate.constraints.append(return_value == 1)

        return [global_state]


    @StateTransition()
    def delegatecall_(self, global_state):
        instr = global_state.get_current_instruction()
        environment = global_state.environment

        try:
            callee_address, callee_account, call_data, _, call_data_type, gas, _, _ = get_call_parameters(global_state,
                                                                                                          self.dynamic_loader)
        except ValueError as e:
            logging.info(
                "Could not determine required parameters for call, putting fresh symbol on the stack. \n{}".format(e)
            )
            global_state.mstate.stack.append(global_state.new_bitvec("retval_" + str(instr['address']), 256))
            return [global_state]

        transaction = MessageCallTransaction(global_state.world_state,
                                             environment.active_account,
                                             environment.sender,
                                             call_data,
                                             gas_price=environment.gasprice,
                                             call_value=environment.callvalue,
                                             origin=environment.origin,
                                             call_data_type=call_data_type,
                                             code=callee_account.code
                                             )
        raise TransactionStartSignal(transaction, self.op_code)


    @StateTransition()
    def delegatecall_post(self, global_state):
        instr = global_state.get_current_instruction()

        try:
            _, _, _, _, _, _, memory_out_offset, memory_out_size =\
                get_call_parameters(global_state, self.dynamic_loader)
        except ValueError as e:
            logging.info(
                "Could not determine required parameters for call, putting fresh symbol on the stack. \n{}".format(e)
            )
            global_state.mstate.stack.append(global_state.new_bitvec("retval_" + str(instr['address']), 256))
            return [global_state]

        if global_state.last_return_data is None:
            # Put return value on stack
            return_value = global_state.new_bitvec("retval_" + str(instr['address']), 256)
            global_state.mstate.stack.append(return_value)
            global_state.mstate.constraints.append(return_value == 0)

            return [global_state]

        try:
            memory_out_offset = util.get_concrete_int(memory_out_offset) if isinstance(memory_out_offset,
                                                                                       ExprRef) else memory_out_offset
            memory_out_size = util.get_concrete_int(memory_out_size) if isinstance(memory_out_size,
                                                                                   ExprRef) else memory_out_size
        except TypeError:
            global_state.mstate.stack.append(global_state.new_bitvec("retval_" + str(instr['address']), 256))
            return [global_state]

            # Copy memory
        global_state.mstate.mem_extend(memory_out_offset,
                                       min(memory_out_size, len(global_state.last_return_data)))
        for i in range(min(memory_out_size, len(global_state.last_return_data))):
            global_state.mstate.memory[i + memory_out_offset] = global_state.last_return_data[i]

        # Put return value on stack
        return_value = global_state.new_bitvec("retval_" + str(instr['address']), 256)
        global_state.mstate.stack.append(return_value)
        global_state.mstate.constraints.append(return_value == 1)

        return [global_state]

    @StateTransition()
    def staticcall_(self, global_state):
        # TODO: implement me
        instr = global_state.get_current_instruction()
        global_state.mstate.stack.append(global_state.new_bitvec("retval_" + str(instr['address']), 256))
        return [global_state]




from z3 import BitVec, Extract, Not
from logging import debug

from mythril.disassembler.disassembly import Disassembly
from mythril.laser.ethereum.cfg import Node, Edge, JumpType
from mythril.laser.ethereum.state import CalldataType
from mythril.laser.ethereum.transaction.transaction_models import MessageCallTransaction, ContractCreationTransaction,\
    get_next_transaction_id

def execute_message_call(laser_evm, callee_address):
    """ Executes a message call transaction from all open states """
    open_states = laser_evm.open_states[:]
    del laser_evm.open_states[:]

    for open_world_state in open_states:
        if open_world_state[callee_address].deleted:
            debug("Can not execute dead contract, skipping.")
            continue

        next_transaction_id = get_next_transaction_id()
        transaction = MessageCallTransaction(
            world_state=open_world_state,
            callee_account=open_world_state[callee_address],
            caller=BitVec("caller{}".format(next_transaction_id), 256),
            identifier=next_transaction_id,
            call_data=[],
            gas_price=BitVec("gas_price{}".format(next_transaction_id), 256),
            call_value=BitVec("call_value{}".format(next_transaction_id), 256),
            origin=BitVec("origin{}".format(next_transaction_id), 256),
            call_data_type=CalldataType.SYMBOLIC,
        )
        _setup_global_state_for_execution(laser_evm, transaction)

    laser_evm.exec()


def execute_contract_creation(laser_evm, contract_initialization_code, contract_name=None):
    """ Executes a contract creation transaction from all open states"""
    open_states = laser_evm.open_states[:]
    del laser_evm.open_states[:]

    new_account = laser_evm.world_state.create_account(0, concrete_storage=True, dynamic_loader=None)
    if contract_name:
        new_account.contract_name = contract_name

    for open_world_state in open_states:
        next_transaction_id = get_next_transaction_id()
        transaction = ContractCreationTransaction(
            open_world_state,
            BitVec("creator{}".format(next_transaction_id), 256),
            next_transaction_id,
            new_account,
            Disassembly(contract_initialization_code),
            [],
            BitVec("gas_price{}".format(next_transaction_id), 256),
            BitVec("call_value{}".format(next_transaction_id), 256),
            BitVec("origin{}".format(next_transaction_id), 256),
            CalldataType.SYMBOLIC
        )
        _setup_global_state_for_execution(laser_evm, transaction)
    laser_evm.exec(True)

    return new_account


def _setup_global_state_for_execution(laser_evm, transaction):
    """ Sets up global state and cfg for a transactions execution"""
    global_state = transaction.initial_global_state()
    global_state.transaction_stack.append((transaction, None))

    new_node = Node(global_state.environment.active_account.contract_name)

    laser_evm.nodes[new_node.uid] = new_node
    if transaction.world_state.node:
        laser_evm.edges.append(Edge(transaction.world_state.node.uid, new_node.uid, edge_type=JumpType.Transaction,
                               condition=None))

        global_state.mstate.constraints = transaction.world_state.node.constraints
        new_node.constraints = global_state.mstate.constraints

    global_state.world_state.transaction_sequence.append(transaction)
    global_state.node = new_node
    new_node.states.append(global_state)
    laser_evm.work_list.append(global_state)



"""
This module implements basic symbolic execution search strategies
"""
from random import randrange
from . import BasicSearchStrategy

try:
    from random import choices
except ImportError:

    # This is for supporting python versions < 3.6
    from itertools import accumulate
    from random import random
    from bisect import bisect

    def choices(population, weights=None):
        """
        Returns a random element out of the population based on weight.
        If the relative weights or cumulative weights are not specified,
        the selections are made with equal probability.
        """
        if weights is None:
            return [population[int(random() * len(population))]]
        cum_weights = accumulate(weights)
        return [population[bisect(cum_weights, random()*cum_weights[-1], 0, len(population)-1)]]


class DepthFirstSearchStrategy(BasicSearchStrategy):
    """
    Implements a depth first search strategy
    I.E. Follow one path to a leaf, and then continue to the next one
    """

    def get_strategic_global_state(self):
        return self.work_list.pop()


class BreadthFirstSearchStrategy(BasicSearchStrategy):
    """
    Implements a breadth first search strategy
    I.E. Execute all states of a "level" before continuing
    """

    def get_strategic_global_state(self):
        return self.work_list.pop(0)


class ReturnRandomNaivelyStrategy(BasicSearchStrategy):
    """
    chooses a random state from the worklist with equal likelihood
    """
    def get_strategic_global_state(self):
        if len(self.work_list) > 0:
            return self.work_list.pop(randrange(len(self.work_list)))
        else:
            raise IndexError


class ReturnWeightedRandomStrategy(BasicSearchStrategy):
    """
    chooses a random state from the worklist with likelihood based on inverse proportion to depth
    """

    def get_strategic_global_state(self):
        probability_distribution = [1/(global_state.mstate.depth+1) for global_state in self.work_list]
        return self.work_list.pop(choices(range(len(self.work_list)), probability_distribution)[0])




from z3 import ExprRef

class KeccakFunctionManager:
    def __init__(self):
        self.keccak_expression_mapping = {}

    def is_keccak(self, expression) -> bool:
        return str(expression) in self.keccak_expression_mapping.keys()

    def get_argument(self, expression) -> ExprRef:
        if not self.is_keccak(expression):
            raise ValueError("Expression is not a recognized keccac result")
        return self.keccak_expression_mapping[str(expression)][1]

    def add_keccak(self, expression: ExprRef, argument: ExprRef):
        index = str(expression)
        self.keccak_expression_mapping[index] = (expression, argument)



import logging
from z3 import simplify
import mythril.laser.ethereum.util as util
from mythril.laser.ethereum.state import Account, CalldataType, GlobalState
from mythril.support.loader import DynLoader
import re

"""
This module contains the business logic used by Instruction in instructions.py
to get the necessary elements from the stack and determine the parameters for the new global state.
"""


def get_call_parameters(global_state: GlobalState, dynamic_loader: DynLoader, with_value=False):
    """
    Gets call parameters from global state
    Pops the values from the stack and determines output parameters
    :param global_state: state to look in
    :param dynamic_loader: dynamic loader to use
    :param with_value: whether to pop the value argument from the stack
    :return: callee_account, call_data, value, call_data_type, gas
    """
    gas, to = global_state.mstate.pop(2)
    value = global_state.mstate.pop() if with_value else 0
    memory_input_offset, memory_input_size, memory_out_offset, memory_out_size = global_state.mstate.pop(4)

    callee_address = get_callee_address(global_state, dynamic_loader, to)

    callee_account = None
    call_data, call_data_type = get_call_data(global_state, memory_input_offset, memory_input_size, False)

    if int(callee_address, 16) >= 5 or int(callee_address, 16) == 0:
        call_data, call_data_type = get_call_data(global_state, memory_input_offset, memory_input_size)
        callee_account = get_callee_account(global_state, callee_address, dynamic_loader)

    return callee_address, callee_account, call_data, value, call_data_type, gas, memory_out_offset, memory_out_size


def get_callee_address(global_state:GlobalState, dynamic_loader: DynLoader, symbolic_to_address):
    """
    Gets the address of the callee
    :param global_state: state to look in
    :param dynamic_loader:  dynamic loader to use
    :param symbolic_to_address: The (symbolic) callee address
    :return: Address of the callee
    """
    environment = global_state.environment

    try:
        callee_address = hex(util.get_concrete_int(symbolic_to_address))
    except TypeError:
        logging.debug("Symbolic call encountered")

        match = re.search(r'storage_(\d+)', str(simplify(symbolic_to_address)))
        logging.debug("CALL to: " + str(simplify(symbolic_to_address)))

        if match is None or dynamic_loader is None:
            raise ValueError()

        index = int(match.group(1))
        logging.debug("Dynamic contract address at storage index {}".format(index))

        # attempt to read the contract address from instance storage
        try:
            callee_address = dynamic_loader.read_storage(environment.active_account.address, index)
        # TODO: verify whether this happens or not
        except:
            logging.debug("Error accessing contract storage.")
            raise ValueError

        # testrpc simply returns the address, geth response is more elaborate.
        if not re.match(r"^0x[0-9a-f]{40}$", callee_address):
            callee_address = "0x" + callee_address[26:]

    return callee_address


def get_callee_account(global_state, callee_address, dynamic_loader):
    """
    Gets the callees account from the global_state
    :param global_state: state to look in
    :param callee_address: address of the callee
    :param dynamic_loader: dynamic loader to use
    :return: Account belonging to callee
    """
    environment = global_state.environment
    accounts = global_state.accounts

    try:
        return global_state.accounts[callee_address]
    except KeyError:
        # We have a valid call address, but contract is not in the modules list
        logging.debug("Module with address " + callee_address + " not loaded.")

    if dynamic_loader is None:
        raise ValueError()

    logging.debug("Attempting to load dependency")

    try:
        code = dynamic_loader.dynld(environment.active_account.address, callee_address)
    except Exception:
        logging.debug("Unable to execute dynamic loader.")
        raise ValueError()
    if code is None:
        logging.debug("No code returned, not a contract account?")
        raise ValueError()
    logging.debug("Dependency loaded: " + callee_address)

    callee_account = Account(callee_address, code, callee_address, dynamic_loader=dynamic_loader)
    accounts[callee_address] = callee_account

    return callee_account


def get_call_data(global_state, memory_start, memory_size, pad=True):
    """
    Gets call_data from the global_state
    :param global_state: state to look in
    :param memory_start: Start index
    :param memory_size: Size
    :return: Tuple containing: call_data array from memory or empty array if symbolic, type found
    """
    state = global_state.mstate
    try:
        # TODO: This only allows for either fully concrete or fully symbolic calldata.
        # Improve management of memory and callata to support a mix between both types.
        call_data = state.memory[util.get_concrete_int(memory_start):util.get_concrete_int(memory_start + memory_size)]
        if len(call_data) < 32 and pad:
            call_data += [0] * (32 - len(call_data))
        call_data_type = CalldataType.CONCRETE
        logging.debug("Calldata: " + str(call_data))
    except TypeError:
        logging.info("Unsupported symbolic calldata offset")
        call_data_type = CalldataType.SYMBOLIC
        call_data = []

    return call_data, call_data_type



from mythril.disassembler.disassembly import Disassembly
from mythril.laser.ethereum.state import MachineState, GlobalState, Environment, Account
from mythril.laser.ethereum.instructions import Instruction


def test_codecopy_concrete():
    # Arrange
    active_account = Account("0x0", code= Disassembly("60606040"))
    environment = Environment(active_account, None, None, None, None, None)
    og_state = GlobalState(None, environment, None, MachineState(gas=10000000))

    og_state.mstate.stack = [2, 2, 2]
    instruction = Instruction("codecopy", dynamic_loader=None)

    # Act
    new_state = instruction.evaluate(og_state)[0]

    # Assert
    assert new_state.mstate.memory[2] == 96
    assert new_state.mstate.memory[3] == 64



import logging
from mythril.disassembler.disassembly import Disassembly
from mythril.laser.ethereum.state import GlobalState, Environment, WorldState
from z3 import BitVec
import array

_next_transaction_id = 0


def get_next_transaction_id():
    global _next_transaction_id
    _next_transaction_id += 1
    return _next_transaction_id


class TransactionEndSignal(Exception):
    """ Exception raised when a transaction is finalized"""
    def __init__(self, global_state, revert=False):
        self.global_state = global_state
        self.revert = revert


class TransactionStartSignal(Exception):
    """ Exception raised when a new transaction is started"""
    def __init__(self, transaction, op_code):
        self.transaction = transaction
        self.op_code = op_code


class MessageCallTransaction:
    """ Transaction object models an transaction"""
    def __init__(self,
                 world_state,
                 callee_account,
                 caller,
                 call_data=(),
                 identifier=None,
                 gas_price=None,
                 call_value=None,
                 origin=None,
                 call_data_type=None,
                 code=None
                 ):
        assert isinstance(world_state, WorldState)
        self.id = identifier or get_next_transaction_id()
        self.world_state = world_state
        self.callee_account = callee_account
        self.caller = caller
        self.call_data = call_data
        self.gas_price = BitVec("gasprice{}".format(identifier), 256) if gas_price is None else gas_price
        self.call_value = BitVec("callvalue{}".format(identifier), 256) if call_value is None else call_value
        self.origin = BitVec("origin{}".format(identifier), 256) if origin is None else origin
        self.call_data_type = BitVec("call_data_type{}".format(identifier), 256) if call_data_type is None else call_data_type
        self.code = code
        self.return_data = None

    def initial_global_state(self):
        # Initialize the execution environment
        environment = Environment(
            self.callee_account,
            self.caller,
            self.call_data,
            self.gas_price,
            self.call_value,
            self.origin,
            code=self.code or self.callee_account.code,
            calldata_type=self.call_data_type,
        )

        global_state = GlobalState(self.world_state, environment, None)
        global_state.environment.active_function_name = 'fallback'

        return global_state

    def end(self, global_state, return_data=None, revert=False):
        self.return_data = return_data
        raise TransactionEndSignal(global_state, revert)


class ContractCreationTransaction:
    """ Transaction object models an transaction"""
    def __init__(self,
                 world_state,
                 caller,
                 identifier=None,
                 callee_account=None,
                 code=None,
                 call_data=(),
                 gas_price=None,
                 call_value=None,
                 origin=None,
                 call_data_type=None,
                 ):
        assert isinstance(world_state, WorldState)
        self.id = identifier or get_next_transaction_id()
        self.world_state = world_state
        # TODO: set correct balance for new account
        self.callee_account = callee_account if callee_account else world_state.create_account(0, concrete_storage=True)

        self.caller = caller

        self.gas_price = BitVec("gasprice{}".format(identifier), 256) if gas_price is None else gas_price
        self.call_value = BitVec("callvalue{}".format(identifier), 256) if call_value is None else call_value
        self.origin = BitVec("origin{}".format(identifier), 256) if origin is None else origin
        self.call_data_type = BitVec("call_data_type{}".format(identifier), 256) if call_data_type is None else call_data_type

        self.call_data = call_data
        self.origin = origin
        self.code = code
        self.return_data = None

    def initial_global_state(self):
        # Initialize the execution environment
        environment = Environment(
            self.callee_account,
            self.caller,
            self.call_data,
            self.gas_price,
            self.call_value,
            self.origin,
            self.code,
            calldata_type=self.call_data_type,
        )

        global_state = GlobalState(self.world_state, environment, None)
        global_state.environment.active_function_name = 'constructor'

        return global_state

    def end(self, global_state, return_data=None, revert=False):

        if not all([isinstance(element, int) for element in return_data]) or len(return_data) == 0:
            self.return_data = None
            raise TransactionEndSignal(global_state)

        contract_code = bytes.hex(array.array('B', return_data).tostring())

        global_state.environment.active_account.code = Disassembly(contract_code)
        self.return_data = global_state.environment.active_account.address
        assert global_state.environment.active_account.code.instruction_list != []

        raise TransactionEndSignal(global_state, revert=revert)





import logging
import re

from mythril.analysis import solver
from mythril.analysis.ops import *
from mythril.analysis.report import Issue
from mythril.analysis.swc_data import TX_ORDER_DEPENDENCE
from mythril.exceptions import UnsatError

'''
MODULE DESCRIPTION:
This module finds the existance of transaction order dependence vulnerabilities.
The following webpage contains an extensive description of the vulnerability: 
https://consensys.github.io/smart-contract-best-practices/known_attacks/#transaction-ordering-dependence-tod-front-running
'''


def execute(statespace):
    """ Executes the analysis module"""
    logging.debug("Executing module: TOD")

    issues = []

    for call in statespace.calls:
        # Do analysis
        interesting_storages = list(_get_influencing_storages(call))
        changing_sstores = list(_get_influencing_sstores(statespace, interesting_storages))

        # Build issue if necessary
        if len(changing_sstores) > 0:
            node = call.node
            instruction = call.state.get_current_instruction()
            issue = Issue(contract=node.contract_name, function=node.function_name, address=instruction['address'],
                          title="Transaction order dependence", swc_id=TX_ORDER_DEPENDENCE, _type="Warning")

            issue.description = \
                "A possible transaction order dependence vulnerability exists in function {}. The value or " \
                "direction of the call statement is determined from a tainted storage location"\
                .format(node.function_name)
            issues.append(issue)

    return issues


# TODO: move to __init__ or util module
def _get_states_with_opcode(statespace, opcode):
    """ Gets all (state, node) tuples in statespace with opcode"""
    for k in statespace.nodes:
        node = statespace.nodes[k]
        for state in node.states:
            if state.get_current_instruction()["opcode"] == opcode:
                yield state, node


def _dependent_on_storage(expression):
    """ Checks if expression is dependent on a storage symbol and returns the influencing storages"""
    pattern = re.compile(r"storage_[a-z0-9_&^]*[0-9]+")
    return pattern.findall(str(simplify(expression)))


def _get_storage_variable(storage, state):
    """
    Get storage z3 object given storage name and the state
    :param storage: storage name example: storage_0
    :param state: state to retrieve the variable from
    :return: z3 object representing storage
    """
    index = int(re.search('[0-9]+', storage).group())
    try:
        return state.environment.active_account.storage[index]
    except KeyError:
        return None


def _can_change(constraints, variable):
    """ Checks if the variable can change given some constraints """
    _constraints = copy.deepcopy(constraints)
    try:
        model = solver.get_model(_constraints)
    except UnsatError:
        return False
    try:
        initial_value = int(str(model.eval(variable, model_completion=True)))
        return _try_constraints(constraints, [variable != initial_value]) is not None
    except AttributeError:
        return False

def _get_influencing_storages(call):
    """ Examines a Call object and returns an iterator of all storages that influence the call value or direction"""
    state = call.state
    node = call.node

    # Get relevant storages
    to, value = call.to, call.value
    storages = []
    if to.type == VarType.SYMBOLIC:
        storages += _dependent_on_storage(to.val)
    if value.type == VarType.SYMBOLIC:
        storages += _dependent_on_storage(value.val)

    # See if they can change within the constraints of the node
    for storage in storages:
        variable = _get_storage_variable(storage, state)
        can_change = _can_change(node.constraints, variable)
        if can_change:
            yield storage


def _get_influencing_sstores(statespace, interesting_storages):
    """ Gets sstore (state, node) tuples that write to interesting_storages"""
    for sstore_state, node in _get_states_with_opcode(statespace, 'SSTORE'):
        index, value = sstore_state.mstate.stack[-1], sstore_state.mstate.stack[-2]
        try:
            index = util.get_concrete_int(index)
        except TypeError:
            index = str(index)
        if "storage_{}".format(index) not in interesting_storages:
            continue

        yield sstore_state, node


# TODO: remove
def _try_constraints(constraints, new_constraints):
    """
    Tries new constraints
    :return Model if satisfiable otherwise None
    """
    _constraints = copy.deepcopy(constraints)
    for constraint in new_constraints:
        _constraints.append(copy.deepcopy(constraint))
    try:
        model = solver.get_model(_constraints)
        return model
    except UnsatError:
        return None



from z3 import *
from mythril.analysis.ops import *
from mythril.analysis.report import Issue
from mythril.analysis import solver
from mythril.analysis.swc_data import REENTRANCY
import re
import logging


'''
MODULE DESCRIPTION:

Check for call.value()() to external addresses
'''

MAX_SEARCH_DEPTH = 64


def search_children(statespace, node, start_index=0, depth=0, results=None):
    if results is None:
        results = []
    logging.debug("SEARCHING NODE %d", node.uid)

    if depth < MAX_SEARCH_DEPTH:

        n_states = len(node.states)

        if n_states > start_index:

            for j in range(start_index, n_states):
                if node.states[j].get_current_instruction()['opcode'] == 'SSTORE':
                    results.append(node.states[j].get_current_instruction()['address'])

        children = []

        for edge in statespace.edges:
            if edge.node_from == node.uid:
                children.append(statespace.nodes[edge.node_to])

        if len(children):
            for node in children:
                return search_children(statespace, node, depth=depth + 1, results=results)

    return results


calls_visited = []


def execute(statespace):

    issues = []

    for call in statespace.calls:

        state = call.state
        address = state.get_current_instruction()['address']

        if call.type == "CALL":

            logging.info("[EXTERNAL_CALLS] Call to: %s, value = %s, gas = %s" % (str(call.to), str(call.value), str(call.gas)))

            if call.to.type == VarType.SYMBOLIC and (call.gas.type == VarType.CONCRETE and call.gas.val > 2300) or (call.gas.type == VarType.SYMBOLIC and "2300" not in str(call.gas)):

                description = "This contract executes a message call to "

                target = str(call.to)
                user_supplied = False

                if "calldata" in target or "caller" in target:

                    if "calldata" in target:
                        description += "an address provided as a function argument. "
                    else:
                        description += "the address of the transaction sender. "

                    user_supplied = True
                else:
                    m = re.search(r'storage_([a-z0-9_&^]+)', str(call.to))

                    if m:
                        idx = m.group(1)

                        func = statespace.find_storage_write(state.environment.active_account.address, idx)

                        if func:

                            description += \
                                "an address found at storage slot " + str(idx) + ". " + \
                                "This storage slot can be written to by calling the function `" + func + "`. "
                            user_supplied = True

                if user_supplied:

                    description += "Generally, it is not recommended to call user-supplied addresses using Solidity's call() construct. " \
                                   "Note that attackers might leverage reentrancy attacks to exploit race conditions or manipulate this contract's state."

                    issue = Issue(contract=call.node.contract_name, function=call.node.function_name,
                                  address=address, title="Message call to external contract", _type="Warning",
                                  description=description, swc_id=REENTRANCY)

                else:

                    description += "to another contract. Make sure that the called contract is trusted and does not execute user-supplied code."

                    issue = Issue(contract=call.node.contract_name, function=call.node.function_name, address=address,
                                  title="Message call to external contract", _type="Informational",
                                  description=description, swc_id=REENTRANCY)

                issues.append(issue)

                if address not in calls_visited:
                    calls_visited.append(address)

                    logging.debug("[EXTERNAL_CALLS] Checking for state changes starting from " + call.node.function_name)

                    # Check for SSTORE in remaining instructions in current node & nodes down the CFG

                    state_change_addresses = search_children(statespace, call.node, call.state_index + 1, depth=0, results=[])

                    logging.debug("[EXTERNAL_CALLS] Detected state changes at addresses: " + str(state_change_addresses))

                    if len(state_change_addresses):
                        for address in state_change_addresses:
                            description = "The contract account state is changed after an external call. " \
                                          "Consider that the called contract could re-enter the function before this " \
                                          "state change takes place. This can lead to business logic vulnerabilities."

                            issue = Issue(contract=call.node.contract_name, function=call.node.function_name,
                                          address=address, title="State change after external call", _type="Warning",
                                          description=description, swc_id=REENTRANCY)
                            issues.append(issue)

    return issues



from mythril.analysis.report import Issue
from mythril.analysis.swc_data import *
from mythril.laser.ethereum.cfg import JumpType
"""
MODULE DESCRIPTION:

Check for multiple sends in a single transaction
"""


def execute(statespace):
    issues = []

    for call in statespace.calls:
        findings = []
        # explore state
        findings += _explore_states(call, statespace)
        # explore nodes
        findings += _explore_nodes(call, statespace)

        if len(findings) > 0:
            node = call.node
            instruction = call.state.get_current_instruction()
            issue = Issue(contract=node.contract_name, function=node.function_name, address=instruction['address'],
                          swc_id=MULTIPLE_SENDS, title="Multiple Calls", _type="Informational")

            issue.description = \
                "Multiple sends exist in one transaction. Try to isolate each external call into its own transaction," \
                " as external calls can fail accidentally or deliberately.\nConsecutive calls: \n"

            for finding in findings:
                issue.description += \
                    "Call at address: {}\n".format(finding.state.get_current_instruction()['address'])

            issues.append(issue)
    return issues


def _explore_nodes(call, statespace):
    children = _child_nodes(statespace, call.node)
    sending_children = list(filter(lambda c: c.node in children, statespace.calls))
    return sending_children


def _explore_states(call, statespace):
    other_calls = list(
            filter(lambda other: other.node == call.node and other.state_index > call.state_index, statespace.calls)
        )
    return other_calls


def _child_nodes(statespace, node):
    result = []
    children = [statespace.nodes[edge.node_to] for edge in statespace.edges if edge.node_from == node.uid
                and edge.type != JumpType.Transaction]

    for child in children:
        result.append(child)
        result += _child_nodes(statespace, child)

    return result



import logging, copy
import mythril.laser.ethereum.util as helper
from mythril.laser.ethereum.cfg import JumpType

class TaintRecord:
    """
    TaintRecord contains tainting information for a specific (state, node)
    the information specifies the taint status before executing the operation belonging to the state
    """

    def __init__(self):
        """ Builds a taint record """
        self.stack = []
        self.memory = {}
        self.storage = {}
        self.states = []

    def stack_tainted(self, index):
        """ Returns taint value of stack element at index"""
        if index < len(self.stack):
            return self.stack[index]
        return None

    def memory_tainted(self, index):
        """ Returns taint value of memory element at index"""
        if index in self.memory.keys():
            return self.memory[index]
        return False

    def storage_tainted(self, index):
        """ Returns taint value of storage element at index"""
        if index in self.storage.keys():
            return self.storage[index]
        return False

    def add_state(self, state):
        """ Adds state with this taint record """
        self.states.append(state)

    def clone(self):
        """ Clones this record"""
        clone = TaintRecord()
        clone.stack = copy.deepcopy(self.stack)
        clone.memory = copy.deepcopy(self.memory)
        clone.storage = copy.deepcopy(self.storage)
        return clone


class TaintResult:
    """ Taint analysis result obtained after having ran the taint runner"""

    def __init__(self):
        self.records = []

    def check(self, state, stack_index):
        """
        Checks if stack variable is tainted, before executing the instruction
        :param state: state to check variable in
        :param stack_index: index of stack variable
        :return: tainted
        """
        record = self._try_get_record(state)
        if record is None:
            return None
        return record.stack_tainted(stack_index)

    def add_records(self, records):
        """ Adds records to this taint result """
        self.records += records

    def _try_get_record(self, state):
        """ Finds record belonging to the state """
        for record in self.records:
            if state in record.states:
                return record
        return None


class TaintRunner:
    """
    Taint runner, is able to run taint analysis on symbolic execution result
    """

    @staticmethod
    def execute(statespace, node, state, initial_stack=None):
        """
        Runs taint analysis on the statespace
        :param statespace: symbolic statespace to run taint analysis on
        :param node: taint introduction node
        :param state: taint introduction state
        :param stack_indexes: stack indexes to introduce taint
        :return: TaintResult object containing analysis results
        """
        if initial_stack is None:
            initial_stack = []
        result = TaintResult()
        transaction_stack_length = len(node.states[0].transaction_stack)
        # Build initial current_node
        init_record = TaintRecord()
        init_record.stack = initial_stack

        state_index = node.states.index(state)

        # List of (Node, TaintRecord, index)
        current_nodes = [(node, init_record, state_index)]
        environment = node.states[0].environment

        for node, record, index in current_nodes:
            records = TaintRunner.execute_node(node, record, index)

            result.add_records(records)
            if len(records) == 0:          # continue if there is no record to work on
                continue
            children = TaintRunner.children(node, statespace, environment, transaction_stack_length)
            for child in children:
                current_nodes.append((child, records[-1], 0))
        return result

    @staticmethod
    def children(node, statespace, environment, transaction_stack_length):
        direct_children = [statespace.nodes[edge.node_to] for edge in statespace.edges if edge.node_from == node.uid and edge.type != JumpType.Transaction]
        children = []
        for child in direct_children:
            if all(len(state.transaction_stack) == transaction_stack_length for state in child.states):
                children.append(child)
            elif all(len(state.transaction_stack) > transaction_stack_length for state in child.states):
                children += TaintRunner.children(child, statespace, environment, transaction_stack_length)
        return children

    @staticmethod
    def execute_node(node, last_record, state_index=0):
        """
        Runs taint analysis on a given node
        :param node: node to analyse
        :param last_record: last taint record to work from
        :param state_index: state index to start from
        :return: List of taint records linked to the states in this node
        """
        records = [last_record]
        for index in range(state_index, len(node.states)):
            current_state = node.states[index]
            records.append(TaintRunner.execute_state(records[-1], current_state))
        return records[1:]

    @staticmethod
    def execute_state(record, state):
        assert len(state.mstate.stack) == len(record.stack)
        """ Runs taint analysis on a state """
        record.add_state(state)
        new_record = record.clone()

        # Apply Change
        op = state.get_current_instruction()['opcode']

        if op in TaintRunner.stack_taint_table.keys():
            mutator = TaintRunner.stack_taint_table[op]
            TaintRunner.mutate_stack(new_record, mutator)
        elif op.startswith("PUSH"):
            TaintRunner.mutate_push(op, new_record)
        elif op.startswith("DUP"):
            TaintRunner.mutate_dup(op, new_record)
        elif op.startswith("SWAP"):
            TaintRunner.mutate_swap(op, new_record)
        elif op is "MLOAD":
            TaintRunner.mutate_mload(new_record, state.mstate.stack[-1])
        elif op.startswith("MSTORE"):
            TaintRunner.mutate_mstore(new_record, state.mstate.stack[-1])
        elif op is "SLOAD":
            TaintRunner.mutate_sload(new_record, state.mstate.stack[-1])
        elif op is "SSTORE":
            TaintRunner.mutate_sstore(new_record, state.mstate.stack[-1])
        elif op.startswith("LOG"):
            TaintRunner.mutate_log(new_record, op)
        elif op in ('CALL', 'CALLCODE', 'DELEGATECALL', 'STATICCALL'):
            TaintRunner.mutate_call(new_record, op)
        else:
            logging.debug("Unknown operation encountered: {}".format(op))

        return new_record

    @staticmethod
    def mutate_stack(record, mutator):
        pop, push = mutator

        values = []
        for i in range(pop):
            values.append(record.stack.pop())

        taint = any(values)

        for i in range(push):
            record.stack.append(taint)

    @staticmethod
    def mutate_push(op, record):
        TaintRunner.mutate_stack(record, (0, 1))

    @staticmethod
    def mutate_dup(op, record):
        depth = int(op[3:])
        index = len(record.stack) - depth
        record.stack.append(record.stack[index])

    @staticmethod
    def mutate_swap(op, record):
        depth = int(op[4:])
        l = len(record.stack) - 1
        i = l - depth
        record.stack[l], record.stack[i] = record.stack[i], record.stack[l]

    @staticmethod
    def mutate_mload(record, op0):
        _ = record.stack.pop()
        try:
            index = helper.get_concrete_int(op0)
        except TypeError:
            logging.debug("Can't MLOAD taint track symbolically")
            record.stack.append(False)
            return

        record.stack.append(record.memory_tainted(index))

    @staticmethod
    def mutate_mstore(record, op0):
        _, value_taint = record.stack.pop(), record.stack.pop()
        try:
            index = helper.get_concrete_int(op0)
        except TypeError:
            logging.debug("Can't mstore taint track symbolically")
            return

        record.memory[index] = value_taint

    @staticmethod
    def mutate_sload(record, op0):
        _ = record.stack.pop()
        try:
            index = helper.get_concrete_int(op0)
        except TypeError:
            logging.debug("Can't MLOAD taint track symbolically")
            record.stack.append(False)
            return

        record.stack.append(record.storage_tainted(index))

    @staticmethod
    def mutate_sstore(record, op0):
        _, value_taint = record.stack.pop(), record.stack.pop()
        try:
            index = helper.get_concrete_int(op0)
        except TypeError:
            logging.debug("Can't mstore taint track symbolically")
            return

        record.storage[index] = value_taint

    @staticmethod
    def mutate_log(record, op):
        depth = int(op[3:])
        for _ in range(depth + 2):
            record.stack.pop()

    @staticmethod
    def mutate_call(record, op):
        pops = 6
        if op in ('CALL', 'CALLCODE'):
            pops += 1
        for _ in range(pops):
            record.stack.pop()

        record.stack.append(False)

    stack_taint_table = {
        # instruction: (taint source, taint target)
        'POP': (1, 0),
        'ADD': (2, 1),
        'MUL': (2, 1),
        'SUB': (2, 1),
        'AND': (2, 1),
        'OR':  (2, 1),
        'XOR': (2, 1),
        'NOT': (1, 1),
        'BYTE': (2, 1),
        'DIV': (2, 1),
        'MOD': (2, 1),
        'SDIV': (2, 1),
        'SMOD': (2, 1),
        'ADDMOD': (3, 1),
        'MULMOD': (3, 1),
        'EXP': (2, 1),
        'SIGNEXTEND': (2, 1),
        'LT': (2, 1),
        'GT': (2, 1),
        'SLT': (2, 1),
        'SGT': (2, 1),
        'EQ': (2, 1),
        'ISZERO': (1, 1),
        'CALLVALUE': (0, 1),
        'CALLDATALOAD': (1, 1),
        'CALLDATACOPY': (3, 0), #todo
        'CALLDATASIZE': (0, 1),
        'ADDRESS': (0, 1),
        'BALANCE': (1, 1),
        'ORIGIN': (0, 1),
        'CALLER': (0, 1),
        'CODESIZE': (0, 1),
        'SHA3': (2, 1),
        'GASPRICE': (0, 1),
        'CODECOPY': (3, 0),
        'EXTCODESIZE': (1, 1),
        'EXTCODECOPY': (4, 0),
        'RETURNDATASIZE': (0, 1),
        'BLOCKHASH': (1, 1),
        'COINBASE': (0, 1),
        'TIMESTAMP': (0, 1),
        'NUMBER': (0, 1),
        'DIFFICULTY': (0, 1),
        'GASLIMIT': (0, 1),
        'JUMP': (1, 0),
        'JUMPI': (2, 0),
        'PC': (0, 1),
        'MSIZE': (0, 1),
        'GAS': (0, 1),
        'CREATE': (3, 1),
        'RETURN': (2, 0)
    }



import pytest
from mythril.laser.ethereum.state import MachineState
from mythril.laser.ethereum.evm_exceptions import StackUnderflowException

memory_extension_test_data = [
    (0, 0, 10),
    (0, 30, 10),
    (100, 22, 8)
]


@pytest.mark.parametrize("initial_size,start,extension_size", memory_extension_test_data)
def test_memory_extension(initial_size, start, extension_size):
    # Arrange
    machine_state = MachineState(0)
    machine_state.memory = [0] * initial_size

    # Act
    machine_state.mem_extend(start, extension_size)

    # Assert
    assert machine_state.memory_size == len(machine_state.memory)
    assert machine_state.memory_size == max(initial_size, start + extension_size)


stack_pop_too_many_test_data = [
    (0, 1),
    (0, 2),
    (5, 1),
    (5, 10)
]


@pytest.mark.parametrize("initial_size,overflow", stack_pop_too_many_test_data)
def test_stack_pop_too_many(initial_size, overflow):
    # Arrange
    machine_state = MachineState(0)
    machine_state.stack = [42] * initial_size

    # Act + Assert
    with pytest.raises(StackUnderflowException):
        machine_state.pop(initial_size + overflow)


stack_pop_test_data = [
    ([1, 2, 3], 2, [3, 2]),
    ([1, 3, 4, 7, 7, 1, 2], 5, [2, 1, 7, 7, 4])
]


@pytest.mark.parametrize("initial_stack,amount,expected", stack_pop_test_data)
def test_stack_multiple_pop(initial_stack, amount, expected):
    # Arrange
    machine_state = MachineState(0)
    machine_state.stack = initial_stack[:]

    # Act
    results = machine_state.pop(amount)

    # Assert
    assert results == initial_stack[-amount:][::-1]
    assert results == expected
    assert len(machine_state.stack) == len(initial_stack) - amount


def test_stack_multiple_pop_():
    # Arrange
    machine_state = MachineState(0)
    machine_state.stack = [1, 2, 3]

    # Act
    a, b = machine_state.pop(2)

    # Assert
    assert a == 3
    assert b == 2


def test_stack_single_pop():
    # Arrange
    machine_state = MachineState(0)
    machine_state.stack = [1,2,3]

    # Act
    result = machine_state.pop()

    # Assert
    assert isinstance(result, int)


memory_write_test_data = [
    (5, 10, [1, 2, 3]),
    (0, 0, [3, 4]),
    (20, 1, [2, 4, 10])
]


@pytest.mark.parametrize("initial_size, memory_offset, data", memory_write_test_data)
def test_memory_write(initial_size, memory_offset, data):
    # Arrange
    machine_state = MachineState(0)
    machine_state.memory = [0]*initial_size

    # Act
    machine_state.memory_write(memory_offset, data)

    # Assert
    assert len(machine_state.memory) == max(initial_size, memory_offset+len(data))
    assert machine_state.memory[memory_offset:memory_offset+len(data)] == data



from z3 import Z3Exception, simplify
from mythril.laser.ethereum.svm import NodeFlags
import re

colors = [
    {'border': '#26996f', 'background': '#2f7e5b', 'highlight': {'border': '#fff', 'background': '#28a16f'}},
    {'border': '#9e42b3', 'background': '#842899', 'highlight': {'border': '#fff', 'background': '#933da6'}},
    {'border': '#b82323', 'background': '#991d1d', 'highlight': {'border': '#fff', 'background': '#a61f1f'}},
    {'border': '#4753bf', 'background': '#3b46a1', 'highlight': {'border': '#fff', 'background': '#424db3'}},
    {'border': '#26996f', 'background': '#2f7e5b', 'highlight': {'border': '#fff', 'background': '#28a16f'}},
    {'border': '#9e42b3', 'background': '#842899', 'highlight': {'border': '#fff', 'background': '#933da6'}},
    {'border': '#b82323', 'background': '#991d1d', 'highlight': {'border': '#fff', 'background': '#a61f1f'}},
    {'border': '#4753bf', 'background': '#3b46a1', 'highlight': {'border': '#fff', 'background': '#424db3'}},
]


def get_serializable_statespace(statespace):
    nodes = []
    edges = []
    
    color_map = {}
    i = 0
    for k in statespace.accounts:
        color_map[statespace.accounts[k].contract_name] = colors[i]
        i += 1
    
    for node_key in statespace.nodes:

        node = statespace.nodes[node_key]

        code = node.get_cfg_dict()['code']
        code = re.sub("([0-9a-f]{8})[0-9a-f]+", lambda m: m.group(1) + "(...)", code)

        if NodeFlags.FUNC_ENTRY in node.flags:
            code = re.sub("JUMPDEST", node.function_name, code)

        code_split = code.split("\\n")

        truncated_code = code if (len(code_split) < 7) else "\\n".join(code_split[:6]) + "\\n(click to expand +)"

        color = color_map[node.get_cfg_dict()['contract_name']]
                
        def get_state_accounts(node_state):
            state_accounts = []
            for key in node_state.accounts:
                account = node_state.accounts[key].as_dict
                account.pop('code', None)
                account['balance'] = str(account['balance'])
                
                storage = {}
                for storage_key in account['storage']:
                    storage[str(storage_key)] = str(account['storage'][storage_key])
            
                state_accounts.append({
                    'address': key,
                    'storage': storage
                })  
            return state_accounts          
                
        states = [{'machine': x.mstate.as_dict, 'accounts': get_state_accounts(x)} for x in node.states]
        
        for state in states:
            state['machine']['stack'] = [str(s) for s in state['machine']['stack']]
            state['machine']['memory'] = [str(m) for m in state['machine']['memory']]
        
        truncated_code = truncated_code.replace('\\n', '\n')
        code = code.replace('\\n', '\n')
        
        s_node = {
            'id': str(node_key),
            'func': str(node.function_name),
            'label': truncated_code,
            'code': code,
            'truncated': truncated_code,
            'states': states,
            'color': color,
            'instructions': code.split('\n')
        }
        
        nodes.append(s_node)

    for edge in statespace.edges:

        if edge.condition is None:
            label = ""
        else:

            try:
                label = str(simplify(edge.condition)).replace("\n", "")
            except Z3Exception:
                label = str(edge.condition).replace("\n", "")

        label = re.sub("([^_])([\d]{2}\d+)", lambda m: m.group(1) + hex(int(m.group(2))), label)
        code = re.sub("([0-9a-f]{8})[0-9a-f]+", lambda m: m.group(1) + "(...)", code)

        s_edge = {
            'from': str(edge.as_dict['from']),
            'to': str(edge.as_dict['to']),
            'arrows': 'to',
            'label': label,
            'smooth': { 'type': "cubicBezier" }
        }
        
        edges.append(s_edge)

    return {
        'edges': edges,
        'nodes': nodes
    }



import rlp
import binascii
from ethereum.utils import normalize_address, hash32, trie_root, \
    big_endian_int, address, int256, encode_hex, encode_int, \
    big_endian_to_int, int_to_addr, zpad, parse_as_bin, parse_as_int, \
    decode_hex, sha3, is_string, is_numeric
from rlp.sedes import big_endian_int, Binary, binary, CountableList
from ethereum import utils
from ethereum import trie
from ethereum.trie import Trie
from ethereum.securetrie import SecureTrie

BLANK_HASH = utils.sha3(b'')
BLANK_ROOT = utils.sha3rlp(b'')

STATE_DEFAULTS = {
    "txindex": 0,
    "gas_used": 0,
    "gas_limit": 3141592,
    "block_number": 0,
    "block_coinbase": '\x00' * 20,
    "block_difficulty": 1,
    "timestamp": 0,
    "logs": [],
    "receipts": [],
    "bloom": 0,
    "suicides": [],
    "recent_uncles": {},
    "prev_headers": [],
    "refunds": 0,
}


class Account(rlp.Serializable):
    """
    adjusted account from ethereum.state
    """

    fields = [
        ('nonce', big_endian_int),
        ('balance', big_endian_int),
        ('storage', trie_root),
        ('code_hash', hash32)
    ]

    def __init__(self, nonce, balance, storage, code_hash, db, addr):
        self.db = db
        self.address = addr
        super(Account, self).__init__(nonce, balance, storage, code_hash)
        self.storage_cache = {}
        self.storage_trie = SecureTrie(Trie(self.db))
        self.storage_trie.root_hash = self.storage
        self.touched = False
        self.existent_at_start = True
        self._mutable = True
        self.deleted = False

    @property
    def code(self):
        """
        code rlp data
        """
        return self.db.get(self.code_hash)

    def get_storage_data(self, key):
        """
        get storage data
        """
        if key not in self.storage_cache:
            v = self.storage_trie.get(utils.encode_int32(key))
            self.storage_cache[key] = utils.big_endian_to_int(
                rlp.decode(v) if v else b'')
        return self.storage_cache[key]

    @classmethod
    def blank_account(cls, db, addr, initial_nonce=0):
        """
        creates a blank account
        """
        db.put(BLANK_HASH, b'')
        o = cls(initial_nonce, 0, trie.BLANK_ROOT, BLANK_HASH, db, addr)
        o.existent_at_start = False
        return o

    def is_blank(self):
        """
        checks if is a blank account
        """
        return self.nonce == 0 and self.balance == 0 and self.code_hash == BLANK_HASH

class State:
    """
    adjusted state from ethereum.state
    """

    def __init__(self, db, root):
        self.db = db
        self.trie = Trie(self.db, root)
        self.secure_trie = SecureTrie(self.trie)
        self.journal = []
        self.cache = {}

    def get_and_cache_account(self, addr):
        """Gets and caches an account for an addres, creates blank if not found"""

        if addr in self.cache:
            return self.cache[addr]
        rlpdata = self.secure_trie.get(addr)
        if rlpdata == trie.BLANK_NODE and len(addr) == 32: # support for hashed addresses
            rlpdata = self.trie.get(addr)

        if rlpdata != trie.BLANK_NODE:
            o = rlp.decode(rlpdata, Account, db=self.db, address=addr)
        else:
            o = Account.blank_account(
                self.db, addr, 0)
        self.cache[addr] = o
        o._mutable = True
        o._cached_rlp = None
        return o

    def get_all_accounts(self):
        """
        iterates through trie to and yields non-blank leafs as accounts
        """
        for address_hash, rlpdata in self.secure_trie.trie.iter_branch():
            if rlpdata != trie.BLANK_NODE:
                yield rlp.decode(rlpdata, Account, db=self.db, address=address_hash)



class VmException(Exception):
    pass


class StackUnderflowException(IndexError, VmException):
    pass


class StackOverflowException(VmException):
    pass


class InvalidJumpDestination(VmException):
    pass


class InvalidInstruction(VmException):
    pass



from mythril.laser.ethereum.transaction.transaction_models import MessageCallTransaction, ContractCreationTransaction, get_next_transaction_id
from z3 import BitVec
from mythril.laser.ethereum.state import GlobalState, Environment, CalldataType, Account, WorldState
from mythril.disassembler.disassembly import Disassembly
from mythril.laser.ethereum.cfg import Node, Edge, JumpType


def execute_message_call(laser_evm, callee_address, caller_address, origin_address, code, data, gas, gas_price, value):
    """ Executes a message call transaction from all open states """
    open_states = laser_evm.open_states[:]
    del laser_evm.open_states[:]

    for open_world_state in open_states:
        transaction = MessageCallTransaction(
            identifier=get_next_transaction_id(),
            world_state=open_world_state,
            callee_account=open_world_state[callee_address],
            caller=caller_address,
            call_data=data,
            gas_price=gas_price,
            call_value=value,
            origin=origin_address,
            call_data_type=CalldataType.SYMBOLIC,
            code=Disassembly(code)
        )

        _setup_global_state_for_execution(laser_evm, transaction)

    laser_evm.exec()


def _setup_global_state_for_execution(laser_evm, transaction):
    """ Sets up global state and cfg for a transactions execution"""
    global_state = transaction.initial_global_state()
    global_state.transaction_stack.append((transaction, None))

    new_node = Node(global_state.environment.active_account.contract_name)

    laser_evm.nodes[new_node.uid] = new_node
    if transaction.world_state.node:
        laser_evm.edges.append(Edge(transaction.world_state.node.uid, new_node.uid, edge_type=JumpType.Transaction,
                               condition=None))
    global_state.node = new_node
    new_node.states.append(global_state)
    laser_evm.work_list.append(global_state)



from mythril.disassembler.disassembly import Disassembly
from mythril.laser.ethereum import svm
from mythril.laser.ethereum.state import Account
import mythril.laser.ethereum.cfg as cfg


def test_intercontract_call():
    # Arrange
    cfg.gbl_next_uid = 0

    caller_code = Disassembly("6080604052348015600f57600080fd5b5073deadbeefdeadbeefdeadbeefdeadbeefdeadbeef73ffffffffffffffffffffffffffffffffffffffff166389627e13336040518263ffffffff167c0100000000000000000000000000000000000000000000000000000000028152600401808273ffffffffffffffffffffffffffffffffffffffff1673ffffffffffffffffffffffffffffffffffffffff168152602001915050602060405180830381600087803b15801560be57600080fd5b505af115801560d1573d6000803e3d6000fd5b505050506040513d602081101560e657600080fd5b8101908080519060200190929190505050500000a165627a7a72305820fdb1e90f0d9775c94820e516970e0d41380a94624fa963c556145e8fb645d4c90029")
    caller_address = "0xaffeaffeaffeaffeaffeaffeaffeaffeaffeaffe"

    callee_code = Disassembly("608060405260043610603f576000357c0100000000000000000000000000000000000000000000000000000000900463ffffffff16806389627e13146044575b600080fd5b348015604f57600080fd5b506082600480360381019080803573ffffffffffffffffffffffffffffffffffffffff1690602001909291905050506084565b005b8073ffffffffffffffffffffffffffffffffffffffff166108fc3073ffffffffffffffffffffffffffffffffffffffff16319081150290604051600060405180830381858888f1935050505015801560e0573d6000803e3d6000fd5b50505600a165627a7a72305820a6b1335d6f994632bc9a7092d0eaa425de3dea05e015af8a94ad70b3969e117a0029")
    callee_address = "0xdeadbeefdeadbeefdeadbeefdeadbeefdeadbeef"

    caller_account = Account(caller_address, caller_code, contract_name="Caller")
    callee_account = Account(callee_address, callee_code, contract_name="Callee")

    accounts = {
        caller_address: caller_account,
        callee_address: callee_account
    }

    laser = svm.LaserEVM(accounts)

    # Act
    laser.sym_exec(caller_address)

    # Assert
    # Initial node starts in contract caller
    assert len(laser.nodes.keys()) > 0
    assert laser.nodes[0].contract_name == 'Caller'

    # At one point we call into contract callee
    for node in laser.nodes.values():
        if node.contract_name == 'Callee':
            assert len(node.states[0].transaction_stack) > 1
            return

    assert False



import os
from pathlib import PurePath
import re
import sys
import json
import logging
from mythril.ether.ethcontract import ETHContract
from mythril.ether.soliditycontract import SourceMapping
from mythril.exceptions import CriticalError
from mythril.analysis.security import fire_lasers
from mythril.analysis.symbolic import SymExecWrapper
from mythril.analysis.report import Report

from mythril.ether import util
from mythril.laser.ethereum.util import get_instruction_index


def analyze_truffle_project(sigs, args):

    project_root = os.getcwd()

    build_dir = os.path.join(project_root, "build", "contracts")

    files = os.listdir(build_dir)

    for filename in files:

        if re.match(r'.*\.json$', filename) and filename != "Migrations.json":

            with open(os.path.join(build_dir, filename)) as cf:
                contractdata = json.load(cf)

            try:
                name = contractdata['contractName']
                bytecode = contractdata['deployedBytecode']
                filename = PurePath(contractdata['sourcePath']).name
            except KeyError:
                print("Unable to parse contract data. Please use Truffle 4 to compile your project.")
                sys.exit()
            if len(bytecode) < 4:
                continue

            sigs.import_from_solidity_source(contractdata['sourcePath'], solc_args=args.solc_args)
            sigs.write()

            ethcontract = ETHContract(bytecode, name=name)

            address = util.get_indexed_address(0)
            sym = SymExecWrapper(ethcontract, address, args.strategy, max_depth=args.max_depth,
                                 create_timeout=args.create_timeout, execution_timeout=args.execution_timeout)
            issues = fire_lasers(sym)

            if not len(issues):
                if args.outform == 'text' or args.outform == 'markdown':
                    print("# Analysis result for " + name + "\n\nNo issues found.")
                else:
                    result = {'contract': name, 'result': {'success': True, 'error': None, 'issues': []}}
                    print(json.dumps(result))
            else:

                report = Report()
                # augment with source code

                disassembly = ethcontract.disassembly
                source = contractdata['source']

                deployed_source_map = contractdata['deployedSourceMap'].split(";")

                mappings = []

                for item in deployed_source_map:
                    mapping = item.split(":")

                    if len(mapping) > 0 and len(mapping[0]) > 0:
                        offset = int(mapping[0])

                    if len(mapping) > 1 and len(mapping[1]) > 0:
                        length = int(mapping[1])

                    if len(mapping) > 2 and len(mapping[2]) > 0:
                        idx = int(mapping[2])

                    lineno = source.encode('utf-8')[0:offset].count('\n'.encode('utf-8')) + 1

                    mappings.append(SourceMapping(idx, offset, length, lineno))

                for issue in issues:

                    index = get_instruction_index(disassembly.instruction_list, issue.address)

                    if index:
                            try:
                                offset = mappings[index].offset
                                length = mappings[index].length

                                issue.filename = filename
                                issue.code = source.encode('utf-8')[offset:offset + length].decode('utf-8')
                                issue.lineno = mappings[index].lineno
                            except IndexError:
                                logging.debug("No code mapping at index %d", index)

                    report.append_issue(issue)

                if args.outform == 'json':

                    result = {'contract': name, 'result': {'success': True, 'error': None, 'issues': list(map(lambda x: x.as_dict, issues))}}
                    print(json.dumps(result))

                else:
                    if args.outform == 'text':
                        print("# Analysis result for " + name + ":\n\n" + report.as_text())
                    elif args.outform == 'markdown':
                        print(report.as_markdown())



from ethereum import vm, messages, transactions
from ethereum.state import State
from ethereum.slogging import get_logger
from mythril.ether import util
from logging import StreamHandler
from io import StringIO
import re


def trace(code, calldata=""):
    log_handlers = ['eth.vm.op', 'eth.vm.op.stack', 'eth.vm.op.memory', 'eth.vm.op.storage']
    output = StringIO()
    stream_handler = StreamHandler(output)

    for handler in log_handlers:
        log_vm_op = get_logger(handler)
        log_vm_op.setLevel("TRACE")
        log_vm_op.addHandler(stream_handler)

    addr = bytes.fromhex('0123456789ABCDEF0123456789ABCDEF01234567')
    state = State()

    ext = messages.VMExt(state, transactions.Transaction(0, 0, 21000, addr, 0, addr))
    message = vm.Message(addr, addr, 0, 21000, calldata)
    vm.vm_execute(ext, message, util.safe_decode(code))
    stream_handler.flush()
    ret = output.getvalue()
    lines = ret.split("\n")

    state_trace = []
    for line in lines:
        m = re.search(r'pc=b\'(\d+)\'.*op=([A-Z0-9]+)', line)
        if m:
            pc = m.group(1)
            op = m.group(2)
            m = re.match(r'.*stack=(\[.*?\])', line)

            if m:
                stackitems = re.findall(r'b\'(\d+)\'', m.group(1))
                stack = "["

                if len(stackitems):
                    for i in range(0, len(stackitems) - 1):
                        stack += hex(int(stackitems[i])) + ", "
                    stack += hex(int(stackitems[-1]))

                stack += "]"
            else:
                stack = "[]"

            if re.match(r'^PUSH.*', op):
                val = re.search(r'pushvalue=(\d+)', line).group(1)
                pushvalue = hex(int(val))
                state_trace.append({'pc': pc, 'op': op, 'stack': stack, 'pushvalue': pushvalue})
            else:
                state_trace.append({'pc': pc, 'op': op, 'stack': stack})

    return state_trace



import binascii
import rlp
from mythril.ethereum.interface.leveldb.accountindexing import CountableList
from mythril.ethereum.interface.leveldb.accountindexing import ReceiptForStorage, AccountIndexer
import logging
from ethereum import utils
from ethereum.block import BlockHeader, Block
from mythril.ethereum.interface.leveldb.state import State
from mythril.ethereum.interface.leveldb.eth_db import ETH_DB
from mythril.ether.ethcontract import ETHContract
from mythril.exceptions import AddressNotFoundError

# Per https://github.com/ethereum/go-ethereum/blob/master/core/rawdb/schema.go
# prefixes and suffixes for keys in geth
header_prefix = b'h'         # header_prefix + num (uint64 big endian) + hash -> header
body_prefix = b'b'           # body_prefix + num (uint64 big endian) + hash -> block body
num_suffix = b'n'            # header_prefix + num (uint64 big endian) + num_suffix -> hash
block_hash_prefix = b'H'      # block_hash_prefix + hash -> num (uint64 big endian)
block_receipts_prefix = b'r'  # block_receipts_prefix + num (uint64 big endian) + hash -> block receipts
# known geth keys
head_header_key = b'LastBlock'  # head (latest) header hash
# custom prefixes
address_prefix = b'AM'       # address_prefix + hash -> address
# custom keys
address_mapping_head_key = b'accountMapping'  # head (latest) number of indexed block


def _format_block_number(number):
    """
    formats block number to uint64 big endian
    """
    return utils.zpad(utils.int_to_big_endian(number), 8)


def _encode_hex(v):
    """
    encodes hash as hex
    """
    return '0x' + utils.encode_hex(v)


class LevelDBReader(object):
    """
    level db reading interface, can be used with snapshot
    """

    def __init__(self, db):
        self.db = db
        self.head_block_header = None
        self.head_state = None

    def _get_head_state(self):
        """
        gets head state
        """
        if not self.head_state:
            root = self._get_head_block().state_root
            self.head_state = State(self.db, root)
        return self.head_state

    def _get_account(self, address):
        """
        gets account by address
        """
        state = self._get_head_state()
        account_address = binascii.a2b_hex(utils.remove_0x_head(address))
        return state.get_and_cache_account(account_address)

    def _get_block_hash(self, number):
        """
        gets block hash by block number
        """
        num = _format_block_number(number)
        hash_key = header_prefix + num + num_suffix
        return self.db.get(hash_key)

    def _get_head_block(self):
        """
        gets head block header
        """
        if not self.head_block_header:
            block_hash = self.db.get(head_header_key)
            num = self._get_block_number(block_hash)
            self.head_block_header = self._get_block_header(block_hash, num)
            # find header with valid state
            while not self.db.get(self.head_block_header.state_root) and self.head_block_header.prevhash is not None:
                block_hash = self.head_block_header.prevhash
                num = self._get_block_number(block_hash)
                self.head_block_header = self._get_block_header(block_hash, num)

        return self.head_block_header

    def _get_block_number(self, block_hash):
        """Get block number by its hash"""
        number_key = block_hash_prefix + block_hash
        return self.db.get(number_key)

    def _get_block_header(self, block_hash, num):
        """Get block header by block header hash & number"""
        header_key = header_prefix + num + block_hash

        block_header_data = self.db.get(header_key)
        header = rlp.decode(block_header_data, sedes=BlockHeader)
        return header

    def _get_address_by_hash(self, block_hash):
        """Get mapped address by its hash"""
        address_key = address_prefix + block_hash
        return self.db.get(address_key)

    def _get_last_indexed_number(self):
        """Get latest indexed block number"""
        return self.db.get(address_mapping_head_key)

    def _get_block_receipts(self, block_hash, num):
        """Get block transaction receipts by block header hash & number"""
        number = _format_block_number(num)
        receipts_key = block_receipts_prefix + number + block_hash
        receipts_data = self.db.get(receipts_key)
        receipts = rlp.decode(receipts_data, sedes=CountableList(ReceiptForStorage))
        return receipts


class LevelDBWriter(object):
    """
    level db writing interface
    """

    def __init__(self, db):
        self.db = db
        self.wb = None

    def _set_last_indexed_number(self, number):
        """
        sets latest indexed block number
        """
        return self.db.put(address_mapping_head_key, _format_block_number(number))

    def _start_writing(self):
        """
        start writing a batch
        """
        self.wb = self.db.write_batch()

    def _commit_batch(self):
        """
        commit batch
        """
        self.wb.write()

    def _store_account_address(self, address):
        """
        get block transaction receipts by block header hash & number
        """
        address_key = address_prefix + utils.sha3(address)
        self.wb.put(address_key, address)


class EthLevelDB(object):
    """
    Go-Ethereum LevelDB client class
    """

    def __init__(self, path):
        self.path = path
        self.db = ETH_DB(path)
        self.reader = LevelDBReader(self.db)
        self.writer = LevelDBWriter(self.db)

    def get_contracts(self):
        """
        iterate through all contracts
        """
        for account in self.reader._get_head_state().get_all_accounts():
            if account.code is not None:
                code = _encode_hex(account.code)
                contract = ETHContract(code, enable_online_lookup=False)

                yield contract, account.address, account.balance

    def search(self, expression, callback_func):
        """
        searches through all contract accounts
        """
        cnt = 0
        indexer = AccountIndexer(self)

        for contract, address_hash, balance in self.get_contracts():

            if contract.matches_expression(expression):

                try:
                    address = _encode_hex(indexer.get_contract_by_hash(address_hash))
                except AddressNotFoundError:
                    '''
                    The hash->address mapping does not exist in our index. If the index is up-to-date, this likely means
                    that the contract was created by an internal transaction. Skip this contract as right now we don't
                    have a good solution for this.
                    '''

                    continue

                callback_func(contract, address, balance)

            cnt += 1

            if not cnt % 1000:
                logging.info("Searched %d contracts" % cnt)

    def contract_hash_to_address(self, contract_hash):
        """Tries to find corresponding account address"""

        address_hash = binascii.a2b_hex(utils.remove_0x_head(contract_hash))
        indexer = AccountIndexer(self)

        return _encode_hex(indexer.get_contract_by_hash(address_hash))

    def eth_getBlockHeaderByNumber(self, number):
        """
        gets block header by block number
        """
        block_hash = self.reader._get_block_hash(number)
        block_number = _format_block_number(number)
        return self.reader._get_block_header(block_hash, block_number)

    def eth_getBlockByNumber(self, number):
        """
        gets block body by block number
        """
        block_hash = self.reader._get_block_hash(number)
        block_number = _format_block_number(number)
        body_key = body_prefix + block_number + block_hash
        block_data = self.db.get(body_key)
        body = rlp.decode(block_data, sedes=Block)
        return body

    def eth_getCode(self, address):
        """
        gets account code
        """
        account = self.reader._get_account(address)
        return _encode_hex(account.code)

    def eth_getBalance(self, address):
        """
        gets account balance
        """
        account = self.reader._get_account(address)
        return account.balance

    def eth_getStorageAt(self, address, position):
        """
        gets account storage data at position
        """
        account = self.reader._get_account(address)
        return _encode_hex(utils.zpad(utils.encode_int(account.get_storage_data(position)), 32))



from mythril.analysis.report import Report
from mythril.analysis import modules
import pkgutil
import logging


def fire_lasers(statespace, module_names=None):

    issues = []
    _modules = []

    for loader, name, is_pkg in pkgutil.walk_packages(modules.__path__):
        _modules.append(loader.find_module(name).load_module(name))

    logging.info("Starting analysis")

    for module in _modules:
        if not module_names or module.__name__ in module_names:
            logging.info("Executing " + str(module))
            issues += module.execute(statespace)

    return issues



from mythril.analysis import solver
from mythril.analysis.ops import *
from mythril.analysis.report import Issue
from mythril.analysis.swc_data import UNPROTECTED_SELFDESTRUCT
from mythril.exceptions import UnsatError
import logging


'''
MODULE DESCRIPTION:

Check for SUICIDE instructions that either can be reached by anyone, or where msg.sender is checked against a tainted storage index 
(i.e. there's a write to that index is unconstrained by msg.sender).
'''


def execute(state_space):

    logging.debug("Executing module: UNCHECKED_SUICIDE")

    issues = []

    for k in state_space.nodes:
        node = state_space.nodes[k]

        for state in node.states:
            issues += _analyze_state(state, node)

    return issues


def _analyze_state(state, node):
    issues = []
    instruction = state.get_current_instruction()

    if instruction['opcode'] != "SUICIDE":
        return []

    to = state.mstate.stack[-1]

    logging.debug("[UNCHECKED_SUICIDE] suicide in function " + node.function_name)
    description = "The function `" + node.function_name + "` executes the SUICIDE instruction. "

    if "caller" in str(to):
        description += "The remaining Ether is sent to the caller's address.\n"
    elif "storage" in str(to):
        description += "The remaining Ether is sent to a stored address.\n"
    elif "calldata" in str(to):
        description += "The remaining Ether is sent to an address provided as a function argument.\n"
    elif type(to) == BitVecNumRef:
        description += "The remaining Ether is sent to: " + hex(to.as_long()) + "\n"
    else:
        description += "The remaining Ether is sent to: " + str(to) + "\n"

    not_creator_constraints = []
    if len(state.world_state.transaction_sequence) > 1:
        creator = state.world_state.transaction_sequence[0].caller
        for transaction in state.world_state.transaction_sequence[1:]:
            not_creator_constraints.append(Not(Extract(159, 0, transaction.caller) == Extract(159, 0, creator)))
            not_creator_constraints.append(Not(Extract(159, 0, transaction.caller) == 0))

    try:
        model = solver.get_model(node.constraints + not_creator_constraints)

        debug = "SOLVER OUTPUT:\n" + solver.pretty_print_model(model)

        issue = Issue(contract=node.contract_name, function=node.function_name, address=instruction['address'],
                      swc_id=UNPROTECTED_SELFDESTRUCT, title="Unchecked SUICIDE", _type="Warning",
                      description=description, debug=debug)
        issues.append(issue)
    except UnsatError:
            logging.debug("[UNCHECKED_SUICIDE] no model found")

    return issues



from flags import Flags
from enum import Enum

gbl_next_uid = 0  # node counter

class JumpType(Enum):
    CONDITIONAL = 1
    UNCONDITIONAL = 2
    CALL = 3
    RETURN = 4
    Transaction = 5


class NodeFlags(Flags):
    FUNC_ENTRY = 1
    CALL_RETURN = 2


class Node:
    def __init__(self, contract_name, start_addr=0, constraints=None):
        constraints = constraints if constraints else []
        self.contract_name = contract_name
        self.start_addr = start_addr
        self.states = []
        self.constraints = constraints
        self.function_name = "unknown"
        self.flags = NodeFlags()

        # Self-assign a unique ID

        global gbl_next_uid

        self.uid = gbl_next_uid
        gbl_next_uid += 1

    def get_cfg_dict(self):

        code = ""

        for state in self.states:

            instruction = state.get_current_instruction()

            code += str(instruction['address']) + " " + instruction['opcode']
            if instruction['opcode'].startswith("PUSH"):
                code += " " + instruction['argument']

            code += "\\n"

        return dict(contract_name=self.contract_name, start_addr=self.start_addr, function_name=self.function_name,
                    code=code)


class Edge:
    def __init__(self, node_from, node_to, edge_type=JumpType.UNCONDITIONAL, condition=None):
        self.node_from = node_from
        self.node_to = node_to
        self.type = edge_type
        self.condition = condition

    def __str__(self):
        return str(self.as_dict)

    @property
    def as_dict(self):
        return {"from": self.node_from, 'to': self.node_to}



class MythrilBaseException(Exception):
    pass


class CompilerError(MythrilBaseException):
    pass


class UnsatError(MythrilBaseException):
    pass


class NoContractFoundError(MythrilBaseException):
    pass


class CriticalError(MythrilBaseException):
    pass


class AddressNotFoundError(MythrilBaseException):
    pass



class EthJsonRpcError(Exception):
    pass


class ConnectionError(EthJsonRpcError):
    pass


class BadStatusCodeError(EthJsonRpcError):
    pass


class BadJsonError(EthJsonRpcError):
    pass


class BadResponseError(EthJsonRpcError):
    pass



import mythril.laser.ethereum.util as helper
from mythril.ether.ethcontract import ETHContract
from mythril.ether.util import get_solc_json
from mythril.exceptions import NoContractFoundError


class SourceMapping:

    def __init__(self, solidity_file_idx, offset, length, lineno):
        self.solidity_file_idx = solidity_file_idx
        self.offset = offset
        self.length = length
        self.lineno = lineno


class SolidityFile:

    def __init__(self, filename, data):
        self.filename = filename
        self.data = data


class SourceCodeInfo:

    def __init__(self, filename, lineno, code):
        self.filename = filename
        self.lineno = lineno
        self.code = code


def get_contracts_from_file(input_file, solc_args=None):
    data = get_solc_json(input_file, solc_args=solc_args)
    for key, contract in data['contracts'].items():
        filename, name = key.split(":")
        if filename == input_file and len(contract['bin-runtime']):
            yield SolidityContract(input_file, name, solc_args)


class SolidityContract(ETHContract):

    def __init__(self, input_file, name=None, solc_args=None):

        data = get_solc_json(input_file, solc_args=solc_args)

        self.solidity_files = []

        for filename in data['sourceList']:
            with open(filename, 'r', encoding='utf-8') as file:
                code = file.read()
                self.solidity_files.append(SolidityFile(filename, code))

        has_contract = False

        # If a contract name has been specified, find the bytecode of that specific contract
        srcmap_constructor = []
        srcmap = []
        if name:
            for key, contract in sorted(data['contracts'].items()):
                filename, _name = key.split(":")

                if filename == input_file and name == _name and len(contract['bin-runtime']):
                    code = contract['bin-runtime']
                    creation_code = contract['bin']
                    srcmap = contract['srcmap-runtime'].split(";")
                    srcmap_constructor = contract['srcmap'].split(";")
                    has_contract = True
                    break

        # If no contract name is specified, get the last bytecode entry for the input file

        else:
            for key, contract in sorted(data['contracts'].items()):
                filename, name = key.split(":")

                if filename == input_file and len(contract['bin-runtime']):
                    code = contract['bin-runtime']
                    creation_code = contract['bin']
                    srcmap = contract['srcmap-runtime'].split(";")
                    srcmap_constructor = contract['srcmap'].split(";")
                    has_contract = True

        if not has_contract:
            raise NoContractFoundError

        self.mappings = []

        self.constructor_mappings = []

        self._get_solc_mappings(srcmap)
        self._get_solc_mappings(srcmap_constructor, constructor=True)

        super().__init__(code, creation_code, name=name)

    def get_source_info(self, address, constructor=False):
        disassembly = self.creation_disassembly if constructor else self.disassembly
        mappings = self.constructor_mappings if constructor else self.mappings
        index = helper.get_instruction_index(disassembly.instruction_list, address)
        solidity_file = self.solidity_files[mappings[index].solidity_file_idx]

        filename = solidity_file.filename

        offset = mappings[index].offset
        length = mappings[index].length

        code = solidity_file.data.encode('utf-8')[offset:offset + length].decode('utf-8', errors="ignore")
        lineno = mappings[index].lineno

        return SourceCodeInfo(filename, lineno, code)

    def _get_solc_mappings(self, srcmap, constructor=False):
        mappings = self.constructor_mappings if constructor else self.mappings
        for item in srcmap:
            mapping = item.split(":")

            if len(mapping) > 0 and len(mapping[0]) > 0:
                offset = int(mapping[0])

            if len(mapping) > 1 and len(mapping[1]) > 0:
                length = int(mapping[1])

            if len(mapping) > 2 and len(mapping[2]) > 0:
                idx = int(mapping[2])
            lineno = self.solidity_files[idx].data.encode('utf-8')[0:offset].count('\n'.encode('utf-8')) + 1

            mappings.append(SourceMapping(idx, offset, length, lineno))



from ethereum.abi import encode_abi, encode_int
from ethereum.utils import zpad
from ethereum.abi import method_id
from mythril.exceptions import CompilerError
from subprocess import Popen, PIPE
import binascii
import os
import json


def safe_decode(hex_encoded_string):

    if hex_encoded_string.startswith("0x"):
        return bytes.fromhex(hex_encoded_string[2:])
    else:
        return bytes.fromhex(hex_encoded_string)


def get_solc_json(file, solc_binary="solc", solc_args=None):

    cmd = [solc_binary, "--combined-json", "bin,bin-runtime,srcmap,srcmap-runtime"]

    if solc_args:
        cmd.extend(solc_args.split())

    if not "--allow-paths" in cmd:
        cmd.extend(["--allow-paths", "."])
    else:
        for i, arg in enumerate(cmd):
            if arg == "--allow-paths":
                cmd[i + 1] += ",."

    cmd.append(file)

    try:
        p = Popen(cmd, stdout=PIPE, stderr=PIPE)

        stdout, stderr = p.communicate()
        ret = p.returncode

        if ret != 0:
            raise CompilerError("Solc experienced a fatal error (code %d).\n\n%s" % (ret, stderr.decode('UTF-8')))
    except FileNotFoundError:
        raise CompilerError("Compiler not found. Make sure that solc is installed and in PATH, or set the SOLC environment variable.")

    out = stdout.decode("UTF-8")

    if not len(out):
        raise CompilerError("Compilation failed.")

    return json.loads(out)


def encode_calldata(func_name, arg_types, args):
    mid = method_id(func_name, arg_types)
    function_selector = zpad(encode_int(mid), 4)
    args = encode_abi(arg_types, args)
    return "0x" + function_selector.hex() + args.hex()


def get_random_address():
    return binascii.b2a_hex(os.urandom(20)).decode('UTF-8')


def get_indexed_address(index):
    return "0x" + (hex(index)[2:] * 40)


def solc_exists(version):
    solc_binary = os.path.join(os.environ['HOME'], ".py-solc/solc-v" + version, "bin/solc")
    if os.path.exists(solc_binary):
        return True
    else:
        return False



# -*- coding: utf8 -*-

import copy
import hashlib
import logging

from ethereum.utils import ecrecover_to_pub
from py_ecc.secp256k1 import N as secp256k1n
from rlp.utils import ALL_BYTES

from mythril.laser.ethereum.util import bytearray_to_int, sha3


class NativeContractException(Exception):
    pass


def int_to_32bytes(i):   # used because int can't fit as bytes function's input
    o = [0] * 32
    for x in range(32):
        o[31 - x] = i & 0xff
        i >>= 8
    return bytes(o)


def extract32(data, i):
    if i >= len(data):
        return 0
    o = data[i: min(i + 32, len(data))]
    o.extend(bytearray(32 - len(o)))
    return bytearray_to_int(o)


def ecrecover(data):

    try:
        data = bytearray(data)
        v = extract32(data, 32)
        r = extract32(data, 64)
        s = extract32(data, 96)
    except TypeError:
        raise NativeContractException

    message = b''.join([ALL_BYTES[x] for x in data[0:32]])
    if r >= secp256k1n or s >= secp256k1n or v < 27 or v > 28:
        return []
    try:
        pub = ecrecover_to_pub(message, v, r, s)
    except Exception as e:
        logging.info("An error has occured while extracting public key: "+e)
        return []
    o = [0] * 12 + [x for x in sha3(pub)[-20:]]
    return o


def sha256(data):
    try:
        data = bytes(data)
    except TypeError:
        raise NativeContractException
    return hashlib.sha256(data).digest()


def ripemd160(data):
    try:
        data = bytes(data)
    except TypeError:
        raise NativeContractException
    return 12*[0]+[i for i in hashlib.new('ripemd160', data).digest()]


def identity(data):
    return copy.copy(data)


def native_contracts(address, data):
    """
    takes integer address 1, 2, 3, 4
    """

    functions = (ecrecover, sha256, ripemd160, identity)
    return functions[address-1](data)



from mythril.ether import asm, util
from mythril.support.signatures import SignatureDb
import logging


class Disassembly(object):

    def __init__(self, code, enable_online_lookup=False):
        self.instruction_list = asm.disassemble(util.safe_decode(code))
        self.func_hashes = []
        self.func_to_addr = {}
        self.addr_to_func = {}
        self.bytecode = code

        signatures = SignatureDb(enable_online_lookup=enable_online_lookup)  # control if you want to have online sighash lookups
        try:
            signatures.open()  # open from default locations
        except FileNotFoundError:
            logging.info("Missing function signature file. Resolving of function names from signature file disabled.")

        # Parse jump table & resolve function names

        # Need to take from PUSH1 to PUSH4 because solc seems to remove excess 0s at the beginning for optimizing
        jmptable_indices = asm.find_opcode_sequence([("PUSH1", "PUSH2", "PUSH3", "PUSH4"), ("EQ",)],
                                                    self.instruction_list)

        for i in jmptable_indices:
            func_hash = self.instruction_list[i]['argument']

            # Append with missing 0s at the beginning
            func_hash = "0x" + func_hash[2:].rjust(8, "0")

            self.func_hashes.append(func_hash)
            try:
                # tries local cache, file and optional online lookup
                # may return more than one function signature. since we cannot probe for the correct one we'll use the first
                func_names = signatures.get(func_hash)
                if len(func_names) > 1:
                    # ambigious result
                    func_name = "**ambiguous** %s" % func_names[0]  # return first hit but note that result was ambiguous
                else:
                    # only one item
                    func_name = func_names[0]
            except KeyError:
                func_name = "_function_" + func_hash

            try:
                offset = self.instruction_list[i + 2]['argument']
                jump_target = int(offset, 16)

                self.func_to_addr[func_name] = jump_target
                self.addr_to_func[jump_target] = func_name
            except:
                continue

        signatures.write()  # store resolved signatures (potentially resolved online)

    def get_easm(self):
        # todo: tintinweb - print funcsig resolved data from self.addr_to_func?
        return asm.instruction_list_to_easm(self.instruction_list)
